import time
import traceback
from kikit.defs import EDA_TEXT_HJUSTIFY_T, EDA_TEXT_VJUSTIFY_T
import pcbnew
from kikit.panelize_ui_impl import loadPresetChain, obtainPreset, mergePresets
from kikit import panelize_ui
from kikit.panelize import NonFatalErrors, appendItem
from kikit.common import PKG_BASE, findBoardBoundingBox, fromMm
from .common import initDialog, destroyDialog
import kikit.panelize_ui_sections
import wx
import json
import tempfile
import shutil
import os
from threading import Thread
from itertools import chain

PLATFORMS = ["Linux/MacOS", "Windows"]

class ExceptionThread(Thread):
    def run(self):
        self.exception = None
        try:
            super().run()
        except Exception as e:
            self.exception = e
            self.traceback = traceback.format_exc()

def replaceExt(file, ext):
    return os.path.splitext(file)[0] + ext

def pcbnewPythonPath():
    return os.path.dirname(pcbnew.__file__)

def presetDifferential(source, target):
    result = {}
    for sectionName, section in target.items():
        if sectionName not in source:
            result[sectionName] = section
            continue
        updateKeys = {}
        sourceSection = source[sectionName]
        for key, value in section.items():
            if key not in sourceSection or str(sourceSection[key]).lower() != str(value).lower():
                updateKeys[key] = value
        if len(updateKeys) > 0:
            result[sectionName] = updateKeys
    return result


def transplateBoard(source, target, update=lambda x: None):
    CLEAR_MSG = "Clearing the old board in UI"
    RENDER_MSG = "Rendering the new board in UI"

    target.ClearProject()
    target.DeleteAllFootprints()

    items = chain(
        list(target.GetDrawings()),
        list(target.GetFootprints()),
        list(target.GetTracks()),
        list(target.Zones()))
    for x in items:
        update(CLEAR_MSG)
        target.Remove(x)

    for x in list(target.GetNetInfo().NetsByNetcode().values()):
        update(CLEAR_MSG)
        target.Remove(x)

    update(RENDER_MSG)
    target.SetProperties(source.GetProperties())
    update(RENDER_MSG)
    target.SetPageSettings(source.GetPageSettings())
    update(RENDER_MSG)
    target.SetTitleBlock(source.GetTitleBlock())
    for x in source.GetDrawings():
        update(RENDER_MSG)
        appendItem(target, x)
    for x in source.GetFootprints():
        update(RENDER_MSG)
        appendItem(target, x)
    for x in source.GetTracks():
        update(RENDER_MSG)
        appendItem(target, x)
    for x in source.Zones():
        update(RENDER_MSG)
        appendItem(target, x)

    update(RENDER_MSG)
    d = target.GetDesignSettings()
    d.CloneFrom(source.GetDesignSettings())
    target.SetEnabledLayers(source.GetEnabledLayers())



def drawTemporaryNotification(board, sourceFilename):
    try:
        bbox = findBoardBoundingBox(board)
    except Exception:
        # If the output is empty...
        bbox = pcbnew.BOX2I(pcbnew.VECTOR2I(0, 0), pcbnew.VECTOR2I(0, 0))

    lset = board.GetEnabledLayers()
    lset.AddLayer(pcbnew.Margin)
    board.SetEnabledLayers(lset)

    text = pcbnew.PCB_TEXT(board)
    text.SetLayer(pcbnew.Margin)
    text.SetText(f"PREVIEW ONLY. PANEL SAVED IN {sourceFilename}")
    text.SetPosition(pcbnew.VECTOR2I(bbox.GetX() + bbox.GetWidth() // 2, bbox.GetY() + bbox.GetHeight()) + pcbnew.VECTOR2I(0, fromMm(2)))
    text.SetTextThickness(fromMm(0.4))
    text.SetTextSize(pcbnew.VECTOR2I(fromMm(3), fromMm(3)))
    text.SetVertJustify(EDA_TEXT_VJUSTIFY_T.GR_TEXT_VJUSTIFY_TOP)
    text.SetHorizJustify(EDA_TEXT_HJUSTIFY_T.GR_TEXT_HJUSTIFY_CENTER)
    board.Add(text)


class SFile():
    def __init__(self, nameFilter):
        self.nameFilter = nameFilter
        self.description = ""
        self.isGuiRelevant = lambda section: True

    def validate(self, x):
        return x

class SInputFile(SFile):
    def __init__(self, nameFilter):
        super().__init__(nameFilter)
        self.description = "Input file"
        self.isGuiRelevant = lambda section: True

class SOuputFile(SFile):
    def __init__(self, nameFilter):
        super().__init__(nameFilter)
        self.description = "Output file"
        self.isGuiRelevant = lambda section: True

class ParameterWidgetBase:
    def __init__(self, parent, name, parameter):
        self.name = name
        self.parameter = parameter
        self.label = wx.StaticText(parent,
                                   label=name,
                                   size=wx.Size(150, -1),
                                   style=wx.ALIGN_RIGHT)
        self.label.SetToolTip(parameter.description)
        self.fresh = True

    def showIfRelevant(self, preset):
        relevant = self.parameter.isGuiRelevant(preset)
        if self.fresh or self.label.IsShown() != relevant:
            self.label.Show(relevant)
            self.widget.Show(relevant)
            self.fresh = False
            return True
        return False


class TextWidget(ParameterWidgetBase):
    def __init__(self, parent, name, parameter, onChange):
        super().__init__(parent, name, parameter)
        self.widget = wx.TextCtrl(
            parent, wx.ID_ANY, wx.EmptyString, wx.DefaultPosition, wx.DefaultSize, 0)
        self.widget.Bind(wx.EVT_TEXT, onChange)

    def setValue(self, value):
        self.widget.ChangeValue(str(value))

    def getValue(self):
        return self.widget.GetValue()


class ChoiceWidget(ParameterWidgetBase):
    def __init__(self, parent, name, parameter, onChange):
        super().__init__(parent, name, parameter)
        self.widget = wx.Choice(parent, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize,
                                parameter.vals, 0)
        self.widget.SetSelection(0)
        self.widget.Bind(wx.EVT_CHOICE, onChange)

    def setValue(self, value):
        for i, option in enumerate(self.parameter.vals):
            if option.lower() == str(value).lower():
                self.widget.SetSelection(i)
                break

    def getValue(self):
        return self.parameter.vals[self.widget.GetSelection()]


class InputFileWidget(ParameterWidgetBase):
    def __init__(self, parent, name, parameter, onChange):
        super().__init__(parent, name, parameter)
        self.widget = wx.FilePickerCtrl(
            parent, wx.ID_ANY, wx.EmptyString, name,
            parameter.nameFilter, wx.DefaultPosition, wx.DefaultSize,
            wx.FLP_OPEN | wx.FLP_FILE_MUST_EXIST)
        self.widget.Bind(wx.EVT_FILEPICKER_CHANGED, onChange)

    def getValue(self):
        return self.widget.GetPath()

    def setValue(self, value):
        self.widget.SetPath(value)

class OutputFileWidget(ParameterWidgetBase):
    def __init__(self, parent, name, parameter, onChange):
        super().__init__(parent, name, parameter)
        self.widget = wx.FilePickerCtrl(
            parent, wx.ID_ANY, wx.EmptyString, name,
            parameter.nameFilter, wx.DefaultPosition, wx.DefaultSize,
            wx.FLP_SAVE | wx.FLP_OVERWRITE_PROMPT)
        self.widget.Bind(wx.EVT_FILEPICKER_CHANGED, onChange)

    def getValue(self):
        return self.widget.GetPath()

    def setValue(self, value):
        self.widget.SetPath(value)

def obtainParameterWidget(parameter):
    if isinstance(parameter, kikit.panelize_ui_sections.SChoiceBase):
        return ChoiceWidget
    if isinstance(parameter, SInputFile):
        return InputFileWidget
    if isinstance(parameter, SOuputFile):
        return OutputFileWidget
    return TextWidget


class SectionGui():
    def __init__(self, parent, name, section, onResize, onChange):
        self.name = name
        self.parent = parent
        self.container = wx.CollapsiblePane(
            parent, wx.ID_ANY, name, wx.DefaultPosition, wx.DefaultSize,
            wx.CP_DEFAULT_STYLE)
        self.container.Collapse(False)

        self.container.Bind(wx.EVT_COLLAPSIBLEPANE_CHANGED, onResize)
        self.container.SetSizeHints(wx.DefaultSize, wx.DefaultSize)
        self.container.GetPane().SetSizeHints(wx.DefaultSize, wx.DefaultSize)

        self.itemGrid = wx.FlexGridSizer(0, 2, 2, 2)
        self.itemGrid.AddGrowableCol(1)
        self.itemGrid.SetFlexibleDirection(wx.BOTH)
        self.itemGrid.SetNonFlexibleGrowMode(wx.FLEX_GROWMODE_ALL)

        self.items = {
            name: obtainParameterWidget(param)(
                self.container.GetPane(), name, param, onChange)
            for name, param in section.items()
        }
        for widget in self.items.values():
            self.itemGrid.Add(widget.label, 0,  wx.ALL |
                              wx.ALIGN_CENTER_VERTICAL | wx.EXPAND | wx.RIGHT, 5)
            self.itemGrid.Add(widget.widget, 0,  wx.ALL |
                              wx.ALIGN_CENTER_VERTICAL | wx.EXPAND | wx.RIGHT, 5)

        self.container.GetPane().SetSizer(self.itemGrid)

    def populateInitialValue(self, values):
        for name, widget in self.items.items():
            if name not in values:
                continue
            widget.setValue(values[name])

    def collectPreset(self):
        return {name: widget.getValue() for name, widget in self.items.items()}

    def showOnlyRelevantFields(self):
        changed = False
        preset = self.collectPreset()
        for name, widget in self.items.items():
            if name not in preset:
                continue
            ch = widget.showIfRelevant(preset)
            changed = changed or ch
        if changed:
            # This is hacky, but it is the only reliable way to force collapsible
            # pane to correctly adjust its size
            self.container.Collapse()
            self.container.Expand()
        return changed

    def collectReleventPreset(self):
        preset = self.collectPreset()
        return {name: widget.getValue()
                for name, widget in self.items.items()
                if widget.parameter.isGuiRelevant(preset)}


class PanelizeDialog(wx.Dialog):
    def __init__(self, parent=None, board=None, preset=None):
        wx.Dialog.__init__(
            self, parent, title=f'Panelize a board  (version {kikit.__version__})',
            style=wx.DEFAULT_DIALOG_STYLE)
        self.Bind(wx.EVT_CLOSE, self.OnClose, id=self.GetId())

        self.board = board
        self.dirty = False
        self.progressDlg = None
        self.lastPulse = time.time()

        topMostBoxSizer = wx.BoxSizer(wx.VERTICAL)

        middleSizer = wx.BoxSizer(wx.HORIZONTAL)

        maxDisplayArea = wx.Display().GetClientArea()
        self.maxDialogSize = wx.Size(
            min(500, maxDisplayArea.Width),
            min(800, maxDisplayArea.Height - 200))

        self.scrollWindow = wx.ScrolledWindow(
            self, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, wx.VSCROLL)
        self.scrollWindow.SetSizeHints(self.maxDialogSize, wx.Size(self.maxDialogSize.width, -1))
        self.scrollWindow.SetScrollRate(5, 5)
        self._buildSections(self.scrollWindow)
        middleSizer.Add(self.scrollWindow, 0, wx.EXPAND | wx.ALL, 5)

        self._buildOutputSections(middleSizer)

        topMostBoxSizer.Add(middleSizer, 1, wx.EXPAND | wx.ALL, 5)
        self._buildBottomButtons(topMostBoxSizer)

        self.SetSizer(topMostBoxSizer)
        self.populateInitialValue(preset)
        self.buildOutputSections()
        self.showOnlyRelevantFields()
        self.OnResize()

        if os.name != "nt":
            self.SetBackgroundColour( wx.SystemSettings.GetColour(wx.SYS_COLOUR_BACKGROUND))


    def _buildOutputSections(self, sizer):
        internalSizer = wx.BoxSizer(wx.VERTICAL)

        cliLabel = wx.StaticText(self, label="KiKit CLI command:",
                                 size=wx.DefaultSize, style=wx.ALIGN_LEFT)
        internalSizer.Add(cliLabel, 0, wx.EXPAND | wx.ALL, 2)

        self.platformSelector = wx.Choice(self, wx.ID_ANY, wx.DefaultPosition,
            wx.DefaultSize, PLATFORMS, 0)
        if os.name == "nt":
            self.platformSelector.SetSelection(PLATFORMS.index("Windows"))
        else:
            self.platformSelector.SetSelection(0) # Choose posix by default
        self.platformSelector.Bind(wx.EVT_CHOICE, lambda evt: self.buildOutputSections())
        internalSizer.Add(self.platformSelector, 0, wx.EXPAND | wx.ALL, 2 )

        self.kikitCmdWidget = wx.TextCtrl(
            self, wx.ID_ANY, "KiKit Command", wx.DefaultPosition, wx.DefaultSize,
            wx.TE_MULTILINE | wx.TE_READONLY)
        self.kikitCmdWidget.SetSizeHints(
            wx.Size(self.maxDialogSize.width,
                    self.maxDialogSize.height // 2),
            wx.Size(self.maxDialogSize.width, -1))
        cmdFont = self.kikitCmdWidget.GetFont()
        cmdFont.SetFamily(wx.FONTFAMILY_TELETYPE)
        self.kikitCmdWidget.SetFont(cmdFont)
        internalSizer.Add(self.kikitCmdWidget, 0, wx.EXPAND | wx.ALL, 2)

        jsonLabel = wx.StaticText(self, label="KiKit JSON preset (contains only changed keys):",
                                  size=wx.DefaultSize, style=wx.ALIGN_LEFT)
        internalSizer.Add(jsonLabel, 0, wx.EXPAND | wx.ALL, 2)

        self.kikitJsonWidget = wx.TextCtrl(
            self, wx.ID_ANY, "KiKit JSON", wx.DefaultPosition, wx.DefaultSize,
            wx.TE_MULTILINE | wx.TE_READONLY)
        self.kikitJsonWidget.SetSizeHints(
            wx.Size(self.maxDialogSize.width,
                    self.maxDialogSize.height // 2),
            wx.Size(self.maxDialogSize.width, -1))
        cmdFont = self.kikitJsonWidget.GetFont()
        cmdFont.SetFamily(wx.FONTFAMILY_TELETYPE)
        self.kikitJsonWidget.SetFont(cmdFont)
        internalSizer.Add(self.kikitJsonWidget, 0, wx.EXPAND | wx.ALL, 2)

        ieButtonsSizer = wx.BoxSizer(wx.HORIZONTAL)
        ieButtonsSizer.Add((0, 0), 1, wx.EXPAND, 5)

        self.importButton = wx.Button(self, wx.ID_ANY, u"Import JSON configuration",
            wx.DefaultPosition, wx.DefaultSize, 0)
        try:
            self.importButton.SetBitmap(wx.BitmapBundle(wx.ArtProvider.GetBitmap(wx.ART_FILE_OPEN)))
        except:
            self.importButton.SetBitmap(wx.ArtProvider.GetBitmap(wx.ART_FILE_OPEN))
        ieButtonsSizer.Add(self.importButton, 0, wx.ALL, 5)
        self.importButton.Bind(wx.EVT_BUTTON, self.onImport)

        self.exportButton = wx.Button(self, wx.ID_ANY, u"Export JSON configuration",
            wx.DefaultPosition, wx.DefaultSize, 0)
        try:
            self.exportButton.SetBitmap(wx.BitmapBundle(wx.ArtProvider.GetBitmap(wx.ART_FILE_SAVE)))
        except:
            self.exportButton.SetBitmap(wx.ArtProvider.GetBitmap(wx.ART_FILE_SAVE))
        ieButtonsSizer.Add(self.exportButton, 0, wx.ALL, 5)
        self.exportButton.Bind(wx.EVT_BUTTON, self.onExport)

        internalSizer.Add(ieButtonsSizer, 1, wx.EXPAND, 5)

        sizer.Add(internalSizer, 0, wx.EXPAND | wx.ALL, 2)

    def _buildSections(self, parentWindow):
        sectionsSizer = wx.BoxSizer(wx.VERTICAL)

        sections = {
            "Input": {
                "Input file": SInputFile("*.kicad_pcb")
            },
            "Output": {
                "Output file": SOuputFile("*.kicad_pcb")
            }
        }
        sections.update(kikit.panelize_ui_sections.availableSections)

        self.sections = {
            name: SectionGui(parentWindow, name, section,
                             lambda evt: self.OnResize(), lambda evt: self.OnChange())
            for name, section in sections.items()
        }
        for section in self.sections.values():
            sectionsSizer.Add(section.container, 0, wx.ALL | wx.EXPAND, 5)

        parentWindow.SetSizer(sectionsSizer)

    def _buildBottomButtons(self, parentSizer):
        button_box = wx.BoxSizer(wx.HORIZONTAL)
        closeButton = wx.Button(self, label='Close')
        self.Bind(wx.EVT_BUTTON, self.OnClose, id=closeButton.GetId())
        button_box.Add(closeButton, 1, wx.RIGHT, 10)
        self.okButton = wx.Button(self, label='Panelize')
        self.Bind(wx.EVT_BUTTON, self.OnPanelize, id=self.okButton.GetId())
        button_box.Add(self.okButton, 1)

        parentSizer.Add(button_box, 0, wx.ALIGN_RIGHT |
                        wx.LEFT | wx.RIGHT | wx.BOTTOM, 10)

    def OnResize(self):
        self.scrollWindow.GetSizer().Layout()
        self.scrollWindow.Fit()
        self.scrollWindow.FitInside()
        self.GetSizer().Layout()
        self.Fit()

    def OnClose(self, event):
        self.EndModal(0)

    def _updatePanelizationProgress(self, message, force=False):
        self.phase = message
        now = time.time()

        if now - self.lastPulse > 1 / 50 or force:
            self.lastPulse = now
            if self.progressDlg is not None:
                self.progressDlg.Pulse(newmsg=f"Running KiKit: {self.phase}")
            if force:
                self.progressDlg.Refresh()
            wx.GetApp().Yield()

    def _panelizationRoutine(self, tempdir, input, panelFile, preset):
        panelize_ui.doPanelization(input, panelFile, preset)

        # KiCAD 6 does something strange here, so we will load an empty
        # file if we read it directly, but we can always make a copy and
        # read that. Copying a file can be lengthy, so we will copy the
        # file in a thread.
        copyPanelName = os.path.join(tempdir, "panel-copy.kicad_pcb")
        shutil.copy(panelFile, copyPanelName)
        try:
            shutil.copy(replaceExt(panelFile, ".kicad_pro"), replaceExt(copyPanelName, "kicad_pro"))
            shutil.copy(replaceExt(panelFile, ".kicad_prl"), replaceExt(copyPanelName, "kicad_prl"))
        except FileNotFoundError:
            # We don't care if we didn't manage to copy the files
            pass
        self.temporary_panel = pcbnew.LoadBoard(copyPanelName)

    def _pulseWhilePcbnewRefresh(self):
        while not self.refreshDone:
            time.sleep(1/50)
            self._updatePanelizationProgress("Pcbnew is updating the preview")


    def OnPanelize(self, event):
        with tempfile.TemporaryDirectory(prefix="kikit") as dirname:
            try:
                self.progressDlg = wx.ProgressDialog(
                    "Running kikit", f"Running KiKit:",
                    parent=self)
                self._updatePanelizationProgress("Starting up")
                self.progressDlg.Show()

                args = self.kikitArgs()
                preset = obtainPreset([], **args)
                input = self.sections["Input"].items["Input file"].getValue()
                if len(input) == 0:
                    dlg = wx.MessageDialog(
                        None, f"No input file specified", "Error", wx.OK)
                    dlg.ShowModal()
                    dlg.Destroy()
                    return
                panelFile = self.sections["Output"].items["Output file"].getValue()
                if len(panelFile) == 0:
                    dlg = wx.MessageDialog(
                        None, f"No output file specified", "Error", wx.OK)
                    dlg.ShowModal()
                    dlg.Destroy()
                    return
                if os.path.realpath(input) == os.path.realpath(pcbnew.GetBoard().GetFileName()):
                    dlg = wx.MessageDialog(
                        None,
                        f"The file {input} is the same as currently opened board. Cannot continue.\n\n" + \
                         "Please, run the panelization tool when no board is opened in pcbnew.",
                        "Error", wx.OK)
                    dlg.ShowModal()
                    dlg.Destroy()
                    return

                # We run as much as possible in a separate thread to not stall
                # the UI...
                thread = ExceptionThread(target=self._panelizationRoutine,
                                         args=(dirname, input, panelFile, preset))
                thread.daemon = True
                thread.start()
                while True:
                    self._updatePanelizationProgress("Panelization")
                    thread.join(timeout=1 / 50)
                    if not thread.is_alive():
                        break
                if thread.exception:
                    raise thread.exception

                # ...however, transplate board and pcbnew.Refresh has to happen
                # in the main thread
                transplateBoard(self.temporary_panel, self.board, self._updatePanelizationProgress)
                drawTemporaryNotification(self.board, panelFile)
                self._updatePanelizationProgress("Pcbnew will now refresh panel, the UI might freeze", force=True)
                pcbnew.Refresh()
                self._updatePanelizationProgress("Done", force=True)
                self.dirty = True
            except Exception as e:
                dlg = wx.MessageDialog(
                    None, f"Cannot perform:\n\n{e}", "Error", wx.OK)
                dlg.ShowModal()
                dlg.Destroy()
            finally:
                self.progressDlg.Hide()
                self.progressDlg.Destroy()
                self.progressDlg = None

    def populateInitialValue(self, initialPreset=None):
        preset = loadPresetChain([":default"])
        if initialPreset is not None:
            mergePresets(preset, initialPreset)
        for name, section in self.sections.items():
            if name.lower() not in preset:
                continue
            section.populateInitialValue(preset[name.lower()])
        self.buildOutputSections()

    def showOnlyRelevantFields(self):
        changed = False
        for section in self.sections.values():
            sectionChanged = section.showOnlyRelevantFields()
            changed = changed or sectionChanged
        return changed

    def collectPreset(self, includeInput=False):
        preset = loadPresetChain([":default"])
        if includeInput:
            preset["input"] = {}
        for name, section in self.sections.items():
            if name.lower() not in preset:
                continue
            preset[name.lower()].update(section.collectPreset())
        return preset

    def collectReleventPreset(self):
        preset = {}
        for name, section in self.sections.items():
            preset[name.lower()] = section.collectReleventPreset()
        del preset["input"]
        del preset["output"]
        return preset

    def OnChange(self):
        if self.showOnlyRelevantFields():
            self.OnResize()
        self.buildOutputSections()

    def buildOutputSections(self):
        defaultPreset = loadPresetChain([":default"])
        preset = self.collectReleventPreset()
        presetUpdates = presetDifferential(defaultPreset, preset)

        self.kikitJsonWidget.ChangeValue(json.dumps(presetUpdates, indent=4))

        command = self._buildUnixCommand(presetUpdates) \
                    if self.platformSelector.GetSelection() == 0 \
                    else self._buildWindowsCommand(presetUpdates)
        self.kikitCmdWidget.ChangeValue(command)

    def _buildUnixCommand(self, presetUpdates):
        kikitCommand = "kikit panelize \\\n"
        for section, values in presetUpdates.items():
            if len(values) == 0:
                continue
            attrs = "; ".join(
                [f"{key}: {value}" for key, value in values.items()])
            kikitCommand += f"    --{section} '{attrs}' \\\n"

        inputFilename = self.sections["Input"].items["Input file"].getValue()
        if len(inputFilename) == 0:
            inputFilename = "<missingInput>"

        outputFilename = self.sections["Output"].items["Output file"].getValue()
        if len(outputFilename) == 0:
            outputFilename = "<missingOutput>"

        kikitCommand += f"    '{inputFilename}' '{outputFilename}'"

        return kikitCommand

    def _buildWindowsCommand(self, presetUpdates):
        kikitCommand = "kikit panelize^\n"
        for section, values in presetUpdates.items():
            if len(values) == 0:
                continue
            attrs = "; ".join(
                [f"{key}: {value}" for key, value in values.items()])
            kikitCommand += f"    --{section} \"{attrs}\" ^\n"

        inputFilename = self.sections["Input"].items["Input file"].getValue()
        if len(inputFilename) == 0:
            inputFilename = "<missingInput>"

        outputFilename = self.sections["Output"].items["Output file"].getValue()
        if len(outputFilename) == 0:
            outputFilename = "<missingOutput>"

        kikitCommand += f"    \"{inputFilename}\" \"{outputFilename}\""
        return kikitCommand


    def kikitArgs(self):
        defaultPreset = loadPresetChain([":default"])
        preset = self.collectReleventPreset()
        presetUpdates = presetDifferential(defaultPreset, preset)

        args = {}
        for section, values in presetUpdates.items():
            if len(values) == 0:
                continue
            args[section] = values
        return args

    def onExport(self, evt):
        with wx.FileDialog(self, "Export configuration", wildcard="KiKit configurations (*.json)|*.json",
                style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT) as fileDialog:

            if fileDialog.ShowModal() == wx.ID_CANCEL:
                return

            pathname = fileDialog.GetPath()
            try:
                defaultPreset = loadPresetChain([":default"])
                preset = self.collectReleventPreset()
                presetUpdates = presetDifferential(defaultPreset, preset)
                with open(pathname, "w", encoding="utf-8") as file:
                    json.dump(presetUpdates, file, indent=4)
                wx.MessageBox(f"Configuration exported to {pathname}", "Success",
                    style=wx.OK | wx.ICON_INFORMATION, parent=self)
            except IOError as e:
                wx.MessageBox(f"Cannot export to file {pathname}: {e}", "Error",
                    style=wx.OK | wx.ICON_ERROR, parent=self)

    def onImport(self, evt):
        with wx.FileDialog(self, "Open KiKit configuration", wildcard="KiKit configurations (*.json)|*.json",
                       style=wx.FD_OPEN | wx.FD_FILE_MUST_EXIST) as fileDialog:

            if fileDialog.ShowModal() == wx.ID_CANCEL:
                return
            pathname = fileDialog.GetPath()
            try:
                with open(pathname, "r", encoding="utf-8") as file:
                    preset = json.load(file)
                    self.populateInitialValue(preset)
                    self.OnChange()
            except Exception as e:
                wx.MessageBox(f"Cannot load configuration: {e}", "Error",
                    style=wx.OK | wx.ICON_ERROR, parent=self)


class PanelizePlugin(pcbnew.ActionPlugin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.preset = {}
        self.dirty = False

    def defaults(self):
        self.name = "KiKit: Panelize PCB"
        self.category = "KiKit"
        self.description = "Create a panel"
        self.icon_file_name = os.path.join(PKG_BASE, "resources", "graphics", "panelizeIcon_24x24.png")
        self.show_toolbar_button = True

    def Run(self):
        try:
            dialog = None
            if not self.dirty and not pcbnew.GetBoard().IsEmpty():
                dlg = wx.MessageDialog(
                    None,
                    "The currently opened board is not empty and it will be " + \
                    "replaced by the panel. Do you wish to continue?\n\n" + \
                    "Note that the panelization tool is supposed to be invoked from a stand-alone pcbnew instance.",
                    "Confirm",
                    wx.YES_NO | wx.NO_DEFAULT | wx.ICON_QUESTION)
                ret = dlg.ShowModal()
                dlg.Destroy()
                if ret == wx.ID_NO:
                    return
            dialog = initDialog(lambda: PanelizeDialog(None, pcbnew.GetBoard(), self.preset))
            dialog.ShowModal()
            self.preset = dialog.collectPreset(includeInput=True)
            self.dirty = self.dirty or dialog.dirty
        except Exception as e:
            dlg = wx.MessageDialog(
                None, f"Cannot perform: {e}", "Error", wx.OK)
            dlg.ShowModal()
            dlg.Destroy()
        finally:
            destroyDialog(dialog)


plugin = PanelizePlugin

if __name__ == "__main__":
    # Run test dialog
    import json
    app = wx.App()

    dialog = PanelizeDialog()
    dialog.ShowModal()
    print(json.dumps(dialog.collectPreset(True), indent=4))





import click

@click.group()
def export():
    """
    Export KiCAD boards
    """
    pass

@click.command()
@click.argument("boardfile", type=click.Path(dir_okay=False))
@click.argument("outputdir", type=click.Path(file_okay=False), default=None)
def gerber(boardfile, outputdir):
    from kikit.export import gerberImpl
    from kikit.common import fakeKiCADGui
    app = fakeKiCADGui()

    gerberImpl(boardfile, outputdir)

@click.command()
@click.argument("boardfile", type=click.Path(dir_okay=False))
@click.argument("outputdir", type=click.Path(file_okay=False), default=None)
def dxf(boardfile, outputdir):
    """
    Export board edges and pads to DXF.

    If no output dir is specified, use working directory.

    This command is designed for building 3D printed stencils
    """
    from kikit.export import dxfImpl
    from kikit.common import fakeKiCADGui
    app = fakeKiCADGui()

    dxfImpl(boardfile, outputdir)

export.add_command(gerber)
export.add_command(dxf)



import csv
from dataclasses import dataclass
from enum import Enum
import re
from typing import OrderedDict
from kikit.project import KiCADProject
import pcbnew
from math import sin, cos, radians
from kikit.common import *
from kikit.defs import MODULE_ATTR_T
from kikit.drc_ui import ReportLevel
from kikit import drc
from kikit import eeschema, eeschema_v6
from kikit.text import kikitTextVars
import sys

# A user can still supply v5 schematics even when we run v6, therefore,
# we have to load the correct schematics and provide the right getters
def extractComponents(filename):
    if filename.endswith(".kicad_sch"):
        return eeschema_v6.extractComponents(filename)
    if filename.endswith(".sch"):
        return eeschema.extractComponents(filename)
    raise RuntimeError(f"Unknown schematic file type specified: {filename}")

def getUnit(component):
    if isinstance(component, eeschema_v6.Symbol):
        return eeschema_v6.getUnit(component)
    return eeschema.getUnit(component)

def getField(component, field):
    if isinstance(component, eeschema_v6.Symbol):
        return eeschema_v6.getField(component, field)
    return eeschema.getField(component, field)

def getReference(component):
    if isinstance(component, eeschema_v6.Symbol):
        return eeschema_v6.getReference(component)
    return eeschema.getReference(component)

def ensurePassingDrc(board):
    failed = drc.runImpl(board,
        useMm=True,
        ignoreExcluded=True,
        strict=False,
        level=ReportLevel.error,
        yieldViolation=lambda x: print(x))
    if failed:
        print("DRC failed. See report above. No files produced")
        sys.exit(1)

def hasNonSMDPins(footprint):
    for pad in footprint.Pads():
        if pad.GetAttribute() != pcbnew.PAD_ATTRIB_SMD:
            return True
    return False

class FormatError(Exception):
    pass

class FootprintOrientationHandling(Enum):
    KiCad = 0
    MirrorBottom = 1

@dataclass
class CorrectionPattern:
    """Single correction pattern to match a component against."""
    footprint: re.Pattern
    part_id: re.Pattern
    x_correction: float
    y_correction: float
    rotation: float

def layerToSide(layer):
    if layer == pcbnew.F_Cu:
        return "T"
    if layer == pcbnew.B_Cu:
        return "B"
    raise RuntimeError(f"Got component with invalid layer {layer}")

def footprintPosition(footprint, placeOffset, compensation):
    pos = footprint.GetPosition() - placeOffset
    angle = -footprint.GetOrientation().AsRadians()
    x = compensation[0] * cos(angle) - compensation[1] * sin(angle)
    y = compensation[0] * sin(angle) + compensation[1] * cos(angle)
    pos += VECTOR2I(fromMm(x), fromMm(y))
    return pos

def footprintOrientation(footprint, compensation, orientation: FootprintOrientationHandling = FootprintOrientationHandling.KiCad):
    if orientation == FootprintOrientationHandling.KiCad:
        return (footprint.GetOrientation().AsDegrees() + compensation[2]) % 360
    if orientation == FootprintOrientationHandling.MirrorBottom:
        if layerToSide(footprint.GetLayer()) == "B":
            return (180 - footprint.GetOrientation().AsDegrees() + compensation[2]) % 360
        else:
            return (footprint.GetOrientation().AsDegrees() + compensation[2]) % 360

    raise AssertionError("Invalid orientation handling")

def parseCompensation(compensation):
    compParts = compensation.split(";")
    if len(compParts) != 3:
        raise FormatError(f"Invalid format of compensation '{compensation}' – there should be 3 parts, got {len(compParts)}")
    try:
        comps = [float(x) for x in compParts]
        return comps
    except Exception:
        raise FormatError(f"Invalid format of compensation '{compensation}' – items are not numbers") from None

def defaultFootprintX(footprint, placeOffset, compensation):
    # Overwrite when footprint requires mirrored X when components are on the bottom side
    return toMm(footprintPosition(footprint, placeOffset, compensation)[0])

def defaultFootprintY(footprint, placeOffset, compensation):
    return -toMm(footprintPosition(footprint, placeOffset, compensation)[1])

def excludeFromPos(footprint):
    return footprint.GetAttributes() & pcbnew.FP_EXCLUDE_FROM_POS_FILES

def readCorrectionPatterns(filename):
    """
    Read footprint correction pattern file.

    The file should be a CSV file with the following fields:
    - Regexp to match to the footprint
    - Regexp to match to the part id (ignored at the moment)
    - X correction
    - Y correction
    - Rotation
    """
    corrections = OrderedDict()
    correctionPatterns = []
    with open(filename, encoding="utf-8") as csvfile:
        sample = csvfile.read(1024)
        dialect = csv.Sniffer().sniff(sample)
        has_header = csv.Sniffer().has_header(sample)
        csvfile.seek(0)
        reader = csv.reader(csvfile, dialect)
        first = True
        for row in reader:
            if has_header and first:
                first = False
                continue
            correctionPatterns.append(
                CorrectionPattern(
                    re.compile(row[0]),
                    re.compile(row[1]),
                    float(row[2]),
                    float(row[3]),
                    float(row[4]),
                )
            )
    return correctionPatterns

def applyCorrectionPattern(correctionPatterns, footprint):
    # FIXME: part ID is currently ignored
    # GetUniStringLibId returns the full footprint name including the
    # library in the form of "Resistor_SMD:R_0402_1005Metric"
    footprintName = str(footprint.GetFPID().GetUniStringLibId())
    for corpat in correctionPatterns:
        if corpat.footprint.match(footprintName):
            return (corpat.x_correction, corpat.y_correction, corpat.rotation)
    return (0, 0, 0)

def noFilter(footprint):
    return True

def collectPosData(board, correctionFields, posFilter=lambda x : True,
                   footprintX=defaultFootprintX, footprintY=defaultFootprintY, bom=None,
                   correctionFile=None, orientationHandling: FootprintOrientationHandling = FootprintOrientationHandling.KiCad):
    """
    Extract position data of the footprints.

    If the optional BOM contains fields "<FABNAME>_CORRECTION" in format
    '<X>;<Y>;<ROTATION>' these corrections of component origin and rotation are
    added to the position (in millimeters and degrees). Read the XY corrections
    by hovering cursor over the intended origin in footprint editor and mark the
    coordinates.
    """
    if bom is None:
        bom = {}
    else:
        bom = { getReference(comp): comp for comp in bom }

    correctionPatterns = []
    if correctionFile is not None:
        correctionPatterns = readCorrectionPatterns(correctionFile)

    footprints = []
    placeOffset = board.GetDesignSettings().GetAuxOrigin()
    for footprint in board.GetFootprints():
        if excludeFromPos(footprint):
            continue
        if posFilter(footprint) and footprint.GetReference() in bom:
            footprints.append(footprint)
    def getCompensation(footprint):
        if footprint.GetReference() not in bom:
            return 0, 0, 0
        field = None
        for fieldName in correctionFields:
            field = getField(bom[footprint.GetReference()], fieldName)
            if field is not None:
                break
        if field is None or field == "":
            return applyCorrectionPattern(
                correctionPatterns,
                footprint)
        try:
            return parseCompensation(field)
        except FormatError as e:
            raise FormatError(f"{footprint.GetReference()}: {e}")
    return [(footprint.GetReference(),
             footprintX(footprint, placeOffset, getCompensation(footprint)),
             footprintY(footprint, placeOffset, getCompensation(footprint)),
             layerToSide(footprint.GetLayer()),
             footprintOrientation(footprint, getCompensation(footprint), orientationHandling)) for footprint in footprints]

def posDataToFile(posData, filename):
    with open(filename, "w", newline="", encoding="utf-8") as csvfile:
        writer = csv.writer(csvfile)
        writer.writerow(["Designator", "Mid X", "Mid Y", "Layer", "Rotation"])
        for line in sorted(posData, key=lambda x: naturalComponentKey(x[0])):
            line = list(line)
            for i in [1, 2, 4]:
                line[i] = f"{line[i]:.2f}" # Most Fab houses expect only 2 decimal digits
            writer.writerow(line)

def isValidSchPath(filename):
    return os.path.splitext(filename)[1] in [".sch", ".kicad_sch"]

def isValidBoardPath(filename):
    return os.path.splitext(filename)[1] in [".kicad_pcb"]

def ensureValidSch(filename):
    if not isValidSchPath(filename):
        raise RuntimeError(f"The path {filename} is not a valid KiCAD schema file")

def ensureValidBoard(filename):
    if not isValidBoardPath(filename):
        raise RuntimeError(f"The path {filename} is not a valid KiCAD PCB file")

def expandNameTemplate(template: str, filetype: str, board: pcbnew.BOARD) -> str:
    if re.findall(r"\{.*\}", template) == []:
        raise RuntimeError(f"The filename template '{template} must contain at least one variable name")
    textVars = kikitTextVars(board, KiCADProject(board.GetFileName()).textVars)
    try:
        return template.format(filetype, **textVars)
    except KeyError as e:
        raise RuntimeError(f"Unknown variable {e} in --nametemplate: {template}")

def naturalComponentKey(reference: str) -> Tuple[str, int]:
    text, num = splitOnReverse(reference, lambda x: x.isdigit())
    return str(text), int(num)



from dataclasses import dataclass, field
from kikit.sexpr import Atom, parseSexprF
from itertools import islice
import os
from typing import Optional
from copy import deepcopy

class SchematicError(RuntimeError):
    pass

@dataclass
class Symbol:
    uuid: Optional[str] = None
    path: Optional[str] = None
    unit: Optional[int] = None
    lib_id: Optional[str] = None
    in_bom: bool = True
    on_board: bool = True
    dnp: bool = False
    properties: dict = field(default_factory=dict)

@dataclass
class SymbolInstance:
    symbol_path: Optional[str] = None
    path: Optional[str] = None
    reference: Optional[str] = None
    unit: Optional[int] = None
    value: Optional[str] = None
    footprint: Optional[str] = None

def getProperty(sexpr, field):
    for x in islice(sexpr, 1, None):
        if len(x) > 0 and \
            isinstance(x[0], Atom) and x[0].value == "property" and \
            isinstance(x[1], Atom) and x[1].value == field:
            return x[2].value
    return None

def isSymbol(sexpr):
    if isinstance(sexpr, Atom) or len(sexpr) == 0:
        return False
    item = sexpr[0]
    return isinstance(item, Atom) and item.value == "symbol"

def isSymbolInstances(sexpr):
    if isinstance(sexpr, Atom) or len(sexpr) == 0:
        return False
    item = sexpr[0]
    return isinstance(item, Atom) and item.value == "symbol_instances"

def isSheet(sexpr):
    if isinstance(sexpr, Atom) or len(sexpr) == 0:
        return False
    item = sexpr[0]
    return isinstance(item, Atom) and item.value == "sheet"

def isUuid(sexpr):
    if isinstance(sexpr, Atom) or len(sexpr) == 0:
        return False
    item = sexpr[0]
    return isinstance(item, Atom) and item.value == "uuid"

def isPath(sexpr):
    if isinstance(sexpr, Atom) or len(sexpr) == 0:
        return False
    item = sexpr[0]
    return isinstance(item, Atom) and item.value == "path"

def getUuid(sexpr):
    for x in islice(sexpr, 1, None):
        if x and x[0] == "uuid":
            return x[1].value
    return None

def getElement(sexpr, name):
    for x in islice(sexpr, 1, None):
        key = getAttributeKey(x)
        if key is None:
            continue
        if key == name:
            return x
    return None

def getAttributeKey(sexpr):
    if not sexpr:
        return None
    key = sexpr[0]
    if not isinstance(key, Atom):
        return None
    return key.value

def extractSymbol(sexpr, path):
    s = Symbol()
    for x in islice(sexpr, 1, None):
        key = getAttributeKey(x)
        if key is None:
            continue
        if key == "lib_id":
            s.lib_id = x[1].value
        elif key == "lib_id":
            s.unit = int(x[1].value)
        elif key == "uuid":
            s.uuid = x[1].value
            s.path = path + "/" + s.uuid
        elif key == "in_bom":
            s.in_bom = x[1].value == "yes"
        elif key == "on_board":
            s.on_board = x[1].value == "yes"
        elif key == "dnp":
            s.dnp = x[1].value == "yes"
        elif key == "property":
            s.properties[x[1].value] = x[2].value
    return s

def extractSymbolInstance(sexpr, sheetPath):
    i = SymbolInstance()
    seen = False

    def collectInstanceProperties(pathElem):
        nonlocal i, seen
        for x in islice(pathElem, 2, None):
            key = getAttributeKey(x)
            if key is None:
                continue
            seen = True
            if key == "reference":
                i.reference = x[1].value
            elif key == "unit":
                i.unit = int(x[1].value)
            elif key == "value":
                i.value = x[1].value
            elif key == "footprint":
                i.footprint = x[1].value

    for x in islice(sexpr, 1, None):
        key = getAttributeKey(x)
        if key is None:
            continue
        if key == "uuid":
            i.symbol_path = sheetPath + "/" + x[1].value
        elif key == "instances":
            projects = [proj for proj in islice(x, 1, None)
                if getAttributeKey(proj) == "project"]
            for proj in projects:
                paths = [path for path in islice(proj, 1, None)
                    if isPath(path) and sheetPath == path.items[1].value]
                for path in paths:
                    collectInstanceProperties(path)
    return i if seen else None

def extractSymbolInstanceV6(sexpr, path):
    s = SymbolInstance()
    s.symbol_path = path + sexpr[1].value
    for x in islice(sexpr, 2, None):
        key = getAttributeKey(x)
        if key is None:
            continue
        if key == "reference":
            s.reference = x[1].value
        elif key == "unit":
            s.unit = int(x[1].value)
        elif key == "value":
            s.value = x[1].value
        elif key == "footprint":
            s.footprint = x[1].value
    return s

def collectSymbols(filename, path = None):
    """
    Crawl given sheet and return two lists - one with symbols, one with
    symbol instances
    """
    isRoot = path is None
    with open(filename, encoding="utf-8") as f:
        sheetSExpr = parseSexprF(f)
    symbols, instances = [], []
    for item in sheetSExpr.items:
        if isUuid(item) and path is None:
            path = "/" + item.items[1].value
        if isSymbol(item):
            symbols.append(extractSymbol(item, path))
            instance = extractSymbolInstance(item, path)
            if instance is not None:
                instances.append(instance)
            continue
        if isSheet(item):
            f = getProperty(item, "Sheet file")
            if f is None:
                # v7 format
                f = getProperty(item, "Sheetfile")
            if f is None:
                raise SchematicError("Invalid format - no Sheet file")
            uuid = getUuid(item)
            dirname = os.path.dirname(filename)
            if len(dirname) > 0:
                f = dirname + "/" + f
            s, i = collectSymbols(f, path + "/" + uuid)
            symbols += s
            instances += i
            continue
        # v6 contains symbol instances in a top-level sheet in symbol instances
        if isSymbolInstances(item) and isRoot:
            for p in item.items:
                if isPath(p):
                    instances.append(extractSymbolInstanceV6(p, path))
            continue
    return symbols, instances


def getField(component, field):
    return component.properties.get(field, None)

def getUnit(component):
    return component.unit

def getReference(component):
    return component.properties["Reference"]

def extractComponents(filename):
    symbols, instances = collectSymbols(filename)
    symbolsDict = {x.path: x for x in symbols}

    assert len(symbols) == len(instances)

    components = []
    for inst in instances:
        s = deepcopy(symbolsDict[inst.symbol_path])
        # Note that s should be unique, so we can safely modify it
        if inst.reference is not None:
            s.properties["Reference"] = inst.reference
        if inst.value is not None:
            s.properties["Value"] = inst.value
        if inst.footprint is not None:
            s.properties["Footprint"] = inst.footprint
        if inst.unit is not None:
            s.unit = inst.unit
        components.append(s)
    return components



import click
from pathlib import Path
import sys
import os
import json
import glob
import shutil
import subprocess
import tempfile
import markdown2
import pybars
from datetime import datetime
from kikit import export

def resolveTemplatePath(path):
    """
    Return a correct template path:
    - if the path matches a directory relative to working directory and the
      directory contains template.json, return that
    - otherwise treat the path as a name into the default template library.
    If none of those are template directories, raise exception.
    """
    if os.path.exists(os.path.join(path, "template.json")):
        return path
    PKG_BASE = os.path.dirname(__file__)
    TEMPLATES = os.path.join(PKG_BASE, "resources/present/templates")
    if os.path.exists(os.path.join(TEMPLATES, path, "template.json")):
        return os.path.join(TEMPLATES, path)
    raise RuntimeError("'{}' is not a name or a path for existing template. Perhaps you miss template.json in the template?")

def readTemplate(path):
    """
    Resolve template path, read the property file and return a subclass of
    Template which can render the template.
    """
    templateClasses = {
        "HtmlTemplate": HtmlTemplate
    }
    path = resolveTemplatePath(path)
    with open(os.path.join(path, "template.json"), encoding="utf-8") as jsonFile:
        parameters = json.load(jsonFile)
    try:
        tType = parameters["type"]
    except KeyError:
        raise RuntimeError("Invalid template.json - missing 'type'")
    try:
        return templateClasses[tType](path)
    except KeyError:
        raise RuntimeError("Unknown template type '{}'".format(tType))

def copyRelativeTo(sourceTree, sourceFile, outputDir):
    sourceTree = os.path.abspath(sourceTree)
    sourceFile = os.path.abspath(sourceFile)
    relPath = os.path.relpath(sourceFile, sourceTree)
    outputDir = os.path.join(outputDir, os.path.dirname(relPath))
    Path(outputDir).mkdir(parents=True, exist_ok=True)
    shutil.copy(sourceFile, outputDir)

class Template:
    def __init__(self, directory):
        self.directory = directory
        with open(os.path.join(directory, "template.json"), encoding="utf-8") as jsonFile:
            self.parameters = json.load(jsonFile)
        self.extraResources = []
        self.boards = []
        self.name = None
        self.repository = None

    def _copyResources(self, outputDirectory):
        """
        Copy all resource files specified by template.json and further specified
        by addResource to the output directory.
        """
        for pattern in self.parameters["resources"]:
            for path in glob.glob(os.path.join(self.directory, pattern), recursive=True):
                copyRelativeTo(self.directory, path, outputDirectory)
        for pattern in self.extraResources:
            for path in glob.glob(pattern, recursive=True):
                copyRelativeTo(".", path, outputDirectory)

    def addResource(self, resource):
        """
        Add a resources. Resource can be specified by a glob pattern. The files
        are treated relative to current working directory.
        """
        self.extraResources.append(resource)

    def addBoard(self, name, comment, boardfile):
        """
        Add board
        """
        self.boards.append({
            "name": name,
            "comment": comment,
            "source": boardfile
        })

    def _renderBoards(self, outputDirectory):
        """
        Convert all boards to images and gerber exports. Enrich self.boards
        with paths of generated files
        """
        pcbdraw = shutil.which("pcbdraw")
        if not pcbdraw:
            raise RuntimeError("PcbDraw needs to be installed in order to render boards")

        dirPrefix = "boards"
        boardDir = os.path.join(outputDirectory, dirPrefix)
        Path(boardDir).mkdir(parents=True, exist_ok=True)
        for boardDesc in self.boards:
            boardName = os.path.basename(boardDesc["source"]).replace(".kicad_pcb", "")
            boardDesc["front"] = os.path.join(dirPrefix, boardName + "-front.png")
            boardDesc["back"] = os.path.join(dirPrefix, boardName + "-back.png")
            boardDesc["gerbers"] = os.path.join(dirPrefix, boardName + "-gerbers.zip")
            boardDesc["file"] = os.path.join(dirPrefix, boardName + ".kicad_pcb")

            subprocess.check_call([pcbdraw, "plot", "--vcuts=Cmts.User", "--silent", "--side=front", boardDesc["source"],
                os.path.join(outputDirectory, boardDesc["front"])])
            subprocess.check_call([pcbdraw, "plot", "--vcuts=Cmts.User", "--silent", "--side=back", boardDesc["source"],
                os.path.join(outputDirectory, boardDesc["back"])])

            tmp = tempfile.mkdtemp()
            export.gerberImpl(boardDesc["source"], tmp)
            shutil.make_archive(os.path.join(outputDirectory, boardDesc["gerbers"])[:-4], "zip", tmp)
            shutil.rmtree(tmp)

            shutil.copy(boardDesc["source"], os.path.join(outputDirectory, boardDesc["file"]))

    def render(self, outputDirectory):
        self._copyResources(outputDirectory)
        self._renderBoards(outputDirectory)
        self._renderPage(outputDirectory)

    def gitRevision(self):
        """
        Return a git revision string if in git repo, None otherwise
        """
        proc = subprocess.run(["git", "rev-parse", "HEAD"], capture_output=True)
        if proc.returncode:
            return None
        return proc.stdout.decode("utf-8")

    def currentDateTime(self):
        return datetime.now().strftime("%d. %m. %Y %H:%M")

    def setName(self, name):
        self.name = name

    def setRepository(self, rep):
        self.repository = rep


class HtmlTemplate(Template):
    def __init__(self, path):
        super().__init__(path)

    def addDescriptionFile(self, description):
        if not description.endswith(".md"):
            raise RuntimeError("Only markdown descriptions are supported for now")
        self.description = markdown2.markdown_path(description, extras=["fenced-code-blocks", "tables"])

    def _renderPage(self, outputDirectory):
        with open(os.path.join(self.directory, "index.html"), encoding="utf-8") as templateFile:
            template = pybars.Compiler().compile(templateFile.read())
        gitRev = self.gitRevision()
        content = template({
            "repo": self.repository,
            "gitRev": gitRev,
            "gitRevShort": gitRev[:7] if gitRev else None,
            "datetime": self.currentDateTime(),
            "name": self.name,
            "boards": self.boards,
            "description": self.description
        })
        with open(os.path.join(outputDirectory, "index.html"),"w", encoding="utf-8") as outFile:
            outFile.write(content)

def boardpage(outdir, description, board, resource, template, repository, name):
    try:
        Path(outdir).mkdir(parents=True, exist_ok=True)
        template = readTemplate(template)
        template.addDescriptionFile(description)
        template.setRepository(repository)
        template.setName(name)
        for r in resource:
            template.addResource(r)
        for name, comment, file in board:
            template.addBoard(name, comment, file)
        template.render(outdir)
    except Exception as e:
        sys.stderr.write("An error occurred: " + str(e) + "\n")
        sys.exit(1)








# This has nothing to do with KiKit plugins, instead, it is the registration
# routine for Action plugins. However, the original implementation used the name
# 'plugin'. In order not to break the existing PCM installations we import this
# and leave it here for at least a couple of months to ensure that everybody
# upgrades to a new PCM package.
from kikit.actionPlugins import importAllPlugins # type: ignore

from typing import Any, Dict, Iterable

from kikit.panelize import Panel
from kikit.substrate import Substrate
import pcbnew
from shapely.geometry import LineString

Preset = Dict[str, Dict[str, Any]]


class HookPlugin:
    """
    This type of plugin has a number of callbacks that are invoked during the
    panelization process. The plugin can tweak the process by modifying the
    panel. Inherit from this class and override the callbacks listed below.

    The same instance of the plugin object is used for invoking all of the
    callbacks. So you can safely store information between the calls.

    If you want to know the precise order of operation, please refer to the
    function kikit.panelize_ui:doPanelization.
    """
    def __init__(self, userArg: str, board: pcbnew.BOARD,
                 preset: Dict[str, Dict[str, Any]]) -> None:
        """
        The constructor of the hook plugin will always receive a single string
        from the user, the source design and the presets Dictionary.
        """
        self.userArg = userArg
        self.board = board
        self.preset = preset

    def prePanelSetup(self, panel: Panel) -> None:
        """
        This callback is invoked just after a panel instance was created and no
        operations were performed on it.
        """
        pass

    def afterPanelSetup(self, panel: Panel) -> None:
        """
        This callback is invoked after the panel has inherited design setting,
        properties and the title block.
        """
        pass

    def afterLayout(self, panel: Panel, substrates: Iterable[Substrate]) -> None:
        """
        This callback is invoked after the boards are placed in panel and before
        the partition line is constructed. substrates is an iterable of
        individual boards substrates in the panel
        """
        pass

    def afterTabs(self, panel: Panel, tabCuts: Iterable[LineString],
                  backboneCuts: Iterable[LineString]) -> None:
        """
        This callback is invoked after the tabs have been formed.
        """
        pass

    def afterFraming(self, panel: Panel, frameCuts: Iterable[LineString]) -> None:
        """
        This callback is invoked after the frame was build and before any frame
        decorators (cuts, fiducials) were placed.
        """
        pass

    def afterCuts(self, panel: Panel) -> None:
        """
        This callback is invoked after the cuts were rendered.
        """
        pass

    def finish(self, panel: Panel) -> None:
        """
        This callback is invoked after the panel is finished, just before
        debugging information is collected and the panel is saved.
        """
        pass


class LayoutPlugin:
    """
    This type of plugin can create user specified board layouts
    """
    def __init__(self, preset: Preset, userArg: str, netPattern: str,
                 refPattern: str, vspace: int, hspace: int, rotation: int) -> None:
        self.preset = preset
        self.userArg = userArg
        self.netPattern = netPattern
        self.refPattern = refPattern
        self.vspace = vspace
        self.hspace = hspace
        self.rotation = rotation

    def buildLayout(self, panel: Panel, inputFile: str,
                    sourceArea: pcbnew.BOX2I) -> Iterable[Substrate]:
        """
        This function is supposed to build the layout (append the boards to the
        panel) and return an iterable of substrates of these boards.
        """
        raise NotImplementedError("Layout plugin has to define buildLayout")

    def buildPartitionLine(self, panel: Panel, framingSubstrates: Iterable[Substrate]) -> None:
        """
        This function should build the partition line in the panel. It gets an
        iterable of extra substrates that represent soon-to-be frame of the
        panel.
        """
        return panel.buildPartitionLineFromBB(framingSubstrates)

    def buildExtraCuts(self, panel: Panel) -> Iterable[LineString]:
        """
        This function can return extra cuts, e.g., from internal backbone. It
        shouldn't deal with tab cuts.
        """
        return []


class FramingPlugin:
    """
    This type of plugin can build custom framing
    """
    def __init__(self, preset: Preset, userArg: str) -> None:
        self.preset = preset
        self.userArg = userArg

    def buildFraming(self, panel: Panel) -> Iterable[LineString]:
        """
        This function should append frame to the panel and return list of cuts.
        """
        raise NotImplementedError("FramingPlugin has to define buildFraming")

    def buildDummyFramingSubstrates(self, substrates: Iterable[Substrate]) -> Iterable[Substrate]:
        """
        This function should build dummy substrates that emulate the
        soon-to-be-frame. These substrates are used for partition line
        computation.
        """
        raise NotImplementedError("FramingPlugin has to define buildDummyFramingSubstrates")


class TabsPlugin:
    """
    This plugin can make custom tabs. It provides two functions, however, you
    should override only one of them.
    """
    def __init__(self, preset: Preset, userArg: str) -> None:
        self.preset = preset
        self.userArg = userArg

    def buildTabAnnotations(self, panel: Panel) -> None:
        """
        This function should append tabs annotations to the panel. The rendering
        will be handled automatically.
        """
        raise NotImplementedError("Tabs plugin has to provide buildTabAnnotations when it doesn't override buildTabs")

    def buildTabs(self, panel: Panel) -> Iterable[LineString]:
        """
        This function can directly build the tabs. In most cases, you don't have
        to override this and instead, override buildTabAnnotations.
        """
        panel.clearTabsAnnotations()
        self.buildTabAnnotations(panel)
        return panel.buildTabsFromAnnotations(self.preset["tabs"]["fillet"])

class CutsPlugin:
    """
    This plugin renders tabs (LineStrings) into board features. The cuts are
    divided into two types so you can, e.g., inset you tab cuts.
    """
    def __init__(self, preset: Preset, userArg: str) -> None:
        self.preset = preset
        self.userArg = userArg

    def renderTabCuts(self, panel: Panel, cuts: Iterable[LineString]) -> None:
        """
        Render tab cuts into the panel.
        """
        raise NotImplementedError("Cuts plugin has to provide renderTabCuts")

    def renderOtherCuts(self, panel: Panel, cuts: Iterable[LineString]) -> None:
        """
        Render any other type of cuts (frame, backbone, etc.)
        """
        raise NotImplementedError("Cuts plugin has to provide renderOtherCuts")


class ToolingPlugin:
    """
    This plugin places tooling holes on the board frame.
    """
    def __init__(self, preset: Preset, userArg: str) -> None:
        self.preset = preset
        self.userArg = userArg

    def buildTooling(self, panel: Panel) -> None:
        """
        Add  tooling holes
        """
        raise NotImplementedError("Tooling plugin has to provide buildTooling")

class FiducialsPlugin:
    """
    This plugin places fiducials holes on the board frame.
    """
    def __init__(self, preset: Preset, userArg: str) -> None:
        self.preset = preset
        self.userArg = userArg

    def buildFiducials(self, panel: Panel) -> None:
        """
        Add  fiducials
        """
        raise NotImplementedError("Fiducials plugin has to provide buildFiducials")

class TextVariablePlugin:
    """
    This plugin provides text variables the user can use in text fields.
    """
    def __init__(self, board: pcbnew.BOARD) -> None:
        self.board = board

    def variables(self) -> Dict[str, Any]:
        """
        This function should return a dictionary from variable names to their
        values. The values don't have to be strings – it can be anything
        convertible to string. Especially, if calculating of the value is
        expensive, you can use kikit.text.Formatter to postpone the value
        computation to the moment when it is used from user text.
        """
        return {}



from __future__ import annotations
import typing
from typing import Any, Dict, List, Optional, Union, Tuple, Callable, Iterable
from kikit.typing import Box, T, ComparableT
from itertools import islice, chain
from math import isclose
from copy import copy

class Interval:
    """
    Basic interval representation
    """
    def __init__(self, a: float, b: float) -> None:
        self.min = min(a, b)
        self.max = max(a, b)

    def __contains__(self, item: float) -> bool:
        return self.min <= item and item <= self.max

    def intersect(self, other: Interval) -> Optional[Interval]:
        """
        Return a new interval representing the overlap, otherwise None
        """
        if self.min > other.max or other.min > self.max:
            return None
        return Interval(max(self.min, other.min), min(self.max, other.max))

    def nontrivialIntersect(self, other: Interval) -> Optional[Interval]:
        """
        Return a new interval representing the overlap larger than a single
        point, otherwise None
        """
        i = self.intersect(other)
        if i is None or i.trivial():
            return None
        return i

    def trivial(self) -> bool:
        return self.min == self.max

    @property
    def length(self) -> float:
        return self.max - self.min

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Interval) and \
               isclose(self.min, other.min) and isclose(self.max, other.max)

    def __repr__(self) -> str:
        return f"I({self.min}, {self.max})"

    def __str__(self) -> str:
        return f"<{self.min}, {self.max}>"

class IntervalList:
    def __init__(self, intervals: List[Interval]) -> None:
        self.intervals = self._normalize(self._toList(intervals))

    def __eq__(self, other: object) -> bool:
        return isinstance(other, IntervalList) and \
               self.intervals == self._toList(other)

    def __repr__(self) -> str:
        return f"IL[ {', '.join([x.__repr__() for x in self.intervals])} ]"

    def __str__(self) -> str:
        return f"IL[ {', '.join([x.__str__() for x in self.intervals])} ]"

    def trivial(self) -> bool:
        return len(self.intervals) == 0

    @staticmethod
    def _toList(object: Union[Interval, List[Interval], IntervalList]) -> List[Interval]:
        """
        Convert the object into a list of intervals
        """
        if isinstance(object, Interval):
            return [object]
        if isinstance(object, list):
            return object
        if isinstance(object, IntervalList):
            return object.intervals
        raise RuntimeError("Uknown object")

    @staticmethod
    def _normalize(intervals: List[Interval]) -> List[Interval]:
        intervals.sort(key=lambda x: x.min)
        newIntervals: List[Interval] = []
        for b in intervals:
            if b.trivial():
                continue
            if len(newIntervals) == 0:
                newIntervals.append(copy(b))
                continue
            a = newIntervals[-1]
            if b.min <= a.max:
                a.max = max(a.max, b.max)
            else:
                newIntervals.append(copy(b))
        return newIntervals

    @staticmethod
    def _eventList(intervals0: List[Interval], intervals1: List[Interval]) -> List[Tuple[int, float, int]]:
        """
        Build event list; event is a tuple (1/0, x, event). Event +1 means open
        interval, -1 means close interval. The events are sorted by x
        """
        base1 = [[(0, x.min, 1), (0, x.max, -1)] for x in intervals0]
        base2 = [[(1, x.min, 1), (1, x.max, -1)] for x in intervals1]
        l = list(chain(*base1, *base2))
        l.sort(key=lambda x: x[1])
        return l

    def union(self, other: Union[Interval, IntervalList]) -> IntervalList:
        """
        Union this interval with other and return new instance
        """
        return IntervalList(self.intervals + self._toList(other))

    def intersect(self, other: Union[Interval, IntervalList]) -> IntervalList:
        """
        Perform self / other and return new instance
        """
        intervals = []
        visibleIntervals = 0
        for _, p, e in self._eventList(self.intervals, self._toList(other)):
            if visibleIntervals == 1 and e == 1:
                start = p
            if visibleIntervals == 2 and e == -1:
                intervals.append(Interval(start, p))
            visibleIntervals += e
            assert visibleIntervals >= 0 and visibleIntervals <= 2
        return IntervalList(intervals)

    def difference(self, other: Union[Interval, IntervalList]) -> IntervalList:
        """
        Perform self / other and return new instance
        """
        intervals: List[Interval] = []
        aOpen = 0
        bOpen = 0
        for ident, p, e in self._eventList(self.intervals, self._toList(other)):
            if (ident == 0 and e == 1) and bOpen == 0: # Rising A, no B
                start = p
            if (ident == 1 and e == -1) and aOpen == 1: # Falling B, A is on
                start = p
            if (ident == 0 and e == -1) and bOpen == 0: # Falling A, no B
                intervals.append(Interval(start, p))
            if (ident == 1 and e == 1) and aOpen == 1: # Raising B, A is on
                intervals.append(Interval(start, p))
            if ident == 0:
                aOpen += e
            else:
                bOpen += e
            assert aOpen >= 0 and aOpen <= 1
            assert bOpen >= 0 and bOpen <= 1
        return IntervalList(intervals)

class BoxNeighbors:
    """
    Given a set of axially arranged non-overlapping boxes answers the query for
    the closest top, bottom, left and right neighbor.

    Neighbor is defined as the closest box in a given direction that has
    non-empty intersection of their projections in the given direction.
    """
    def __init__(self, boxes: Dict[object, Box]) -> None:
        """
        Given a dictionary id -> box initializes the structure.

        Boxes are represented by a tuple (minx, miny, maxx, maxy)
        """
        xProj: Callable[[Box], Interval] = lambda b: Interval(b[0], b[2])
        yProj: Callable[[Box], Interval] = lambda b: Interval(b[1], b[3])

        leftList = self._prepareProjection(yProj, lambda b: -b[2], boxes)
        self._leftQ = self._computeQuery(leftList)
        rightList = self._prepareProjection(yProj, lambda b: b[0], boxes)
        self._rightQ = self._computeQuery(rightList)
        topList = self._prepareProjection(xProj, lambda b: -b[3], boxes)
        self._topQ = self._computeQuery(topList)
        bottomList = self._prepareProjection(xProj, lambda b: b[1], boxes)
        self._bottomQ = self._computeQuery(bottomList)

    @staticmethod
    def _prepareProjection(getInterval: Callable[[Box], Interval],
                           getDistance: Callable[[Box], float],
                           boxes: Dict[object, Box]) -> List[Tuple[object, Interval, float]]:
        x = [(ident, getInterval(b), getDistance(b)) for ident, b in boxes.items()]
        x.sort(key=lambda t: t[2])
        return x

    @staticmethod
    def _computeQuery(list: List[Tuple[object, Interval, float]]) -> Dict[object, List[Tuple[object, IntervalList]]]:
        neighbors = {}
        for i, (ident, interval, pos) in enumerate(list):
            n: List[Tuple[object, IntervalList]] = []
            rest = IntervalList([interval])
            for j in range(i + 1, len(list)):
                nIdent, nInterval, nPos = list[j]
                shadow = rest.intersect(nInterval)
                if shadow.trivial():
                    continue
                n.append((nIdent, shadow))
                rest = rest.difference(nInterval)
                if rest.trivial():
                    break
            neighbors[ident] = n
        return neighbors

    @staticmethod
    def _simplify(result: List[Tuple[object, Any]]) -> List[object]:
        return [ident for ident, _ in result]

    def left(self, ident: object) -> List[object]:
        return self._simplify(self._leftQ[ident])

    def leftC(self, ident: object) -> List[Tuple[object, IntervalList]]:
        return self._leftQ[ident]

    def right(self, ident: object) -> List[object]:
        return self._simplify(self._rightQ[ident])

    def rightC(self, ident: object) -> List[Tuple[object, IntervalList]]:
        return self._rightQ[ident]

    def top(self, ident: object) -> List[object]:
        return self._simplify(self._topQ[ident])

    def topC(self, ident: object) -> List[Tuple[object, IntervalList]]:
        return self._topQ[ident]

    def bottom(self, ident: object) -> List[object]:
        return self._simplify(self._bottomQ[ident])

    def bottomC(self, ident: object) -> List[Tuple[object, IntervalList]]:
        return self._bottomQ[ident]


class AxialLine(Interval):
    """
    Representation of a horizontal or vertical line
    """
    def __init__(self, x: float, y1: float, y2: float, tag: Optional[Any]=None) -> None:
        super().__init__(y1, y2)
        self.x = x
        self.tag = tag

    def cut(self, y: float) -> List[AxialLine]:
        """
        Cut the line at y. Return a list of newly created AxialLines
        """
        if y not in self or y == self.min or y == self.max:
            return [self]
        return [
            AxialLine(self.x, self.min, y, self.tag),
            AxialLine(self.x, y, self.max, self.tag)
        ]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, AxialLine) and \
               isclose(self.x, other.x) and super().__eq__(other)

    def __repr__(self) -> str:
        return f"Line[{self.tag}]({self.x}, {self.min}, {self.max})"

    def __hash__(self) -> int:
        return hash((self.x, self.min, self.max, self.tag))

class ShadowLine:
    """
    Represents a horizontal or vertical line with a shadow (possible
    prolongation)
    """
    def __init__(self, line: AxialLine, shadow: Interval) -> None:
        assert isinstance(line, AxialLine)
        assert isinstance(shadow, Interval)
        self.line = line
        self.shadow = shadow

    @property
    def shadowLine(self) -> AxialLine:
        return AxialLine(self.line.x, self.shadow.min, self.shadow.max,
            self.line.tag)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ShadowLine) and \
               self.line == other.line and self.shadow == other.shadow

    def __repr__(self) -> str:
        return f"Shadow({self.line.__repr__()}, {self.shadow.__repr__()})"

def collectBoxEdges(box: Box) -> Tuple[List[AxialLine], List[AxialLine]]:
    """
    Given a box, return a tuple (horiz edges, vert edges) as lists of AxialLine
    """
    return (
        [AxialLine(box[1], box[0], box[2]), AxialLine(box[3], box[0], box[2])],
        [AxialLine(box[0], box[1], box[3]), AxialLine(box[2], box[1], box[3])]
    )

def collectHardStops(boxes: Iterable[Box]) -> Tuple[List[AxialLine], List[AxialLine]]:
    """
    Given an iterable of boxes, return all partition lines hard stops - i.e.,
    union of all edges as a tuple (horiz edges, vert edges) as lists of
    AxialLine
    """
    from kikit.common import shpBBoxMerge

    hedges, vedges = set(), set()
    commonBox: Optional[Box] = None
    for b in boxes:
        if commonBox is None:
            commonBox = b
        else:
            commonBox = shpBBoxMerge(commonBox, b)
        h, v = collectBoxEdges(b)
        hedges.update(h)
        vedges.update(v)
    assert commonBox is not None
    h, v = collectBoxEdges(commonBox)
    hedges.update(h)
    vedges.update(v)
    return list(hedges), list(vedges)

def defaultSeedFilter(boxIdA: object, boxIdB: object, vertical: bool, seedline: AxialLine) -> bool:
    return True

def collectSeedLines(boxes: Dict[object, Box], seedFilter: Callable[[object, object, bool, AxialLine], bool]) \
        -> Tuple[List[AxialLine], List[AxialLine]]:
    """
    Given a dictionary ident -> box return a list of all midlines between
    neighboring boxes.

    The function seedFilter of type(boxAId, boxBId, vertical, seedLine) -> bool,
    serves as a predicate that can filter unwanted seed lines - e.g., too far
    apart or comming from ghost boxes.

    Returns (horlines, verlines), where the lines are tagged with ident
    """
    neighbors = BoxNeighbors(boxes)
    horlines: List[AxialLine] = []
    verlines: List[AxialLine] = []
    for identA, boxA in boxes.items():
        for identB, shadow in neighbors.leftC(identA):
            mid = (boxA[0] + boxes[identB][2]) / 2
            candidates = [AxialLine(mid, e.min, e.max, identA)
                for e in shadow.intervals]
            verlines.extend([x for x in candidates
                if seedFilter(identA, identB, True, x)])
        for identB, shadow in neighbors.rightC(identA):
            mid = (boxA[2] + boxes[identB][0]) / 2
            candidates = [AxialLine(mid, e.min, e.max, identA)
                for e in shadow.intervals]
            verlines.extend([x for x in candidates
                if seedFilter(identA, identB, True, x)])
        for identB, shadow in neighbors.topC(identA):
            mid = (boxA[1] + boxes[identB][3]) / 2
            candidates = [AxialLine(mid, e.min, e.max, identA)
                for e in shadow.intervals]
            horlines.extend([x for x in candidates
                if seedFilter(identA, identB, False, x)])
        for identB, shadow in neighbors.bottomC(identA):
            mid = (boxA[3] + boxes[identB][1]) / 2
            candidates = [AxialLine(mid, e.min, e.max, identA)
                for e in shadow.intervals]
            horlines.extend([x for x in candidates
                if seedFilter(identA, identB, False, x)])
    return horlines, verlines

def upperBound(sortedCollection: List[T], item: ComparableT, key: Callable[[T], ComparableT]) -> int:
    """
    Given a sorted collection, perform binary search to find element x for which
    the following holds: item < key(x) and the value key(x) is the smallest.
    Returns index of such an element.
    """
    lo = 0
    hi = len(sortedCollection)
    while lo < hi:
        mid = (lo + hi) // 2
        if item < key(sortedCollection[mid]):
            hi = mid
        else:
            lo = mid + 1
    return lo

def lowerBound(sortedCollection: List[T], item: ComparableT, key: Callable[[T], ComparableT]) -> int:
    """
    Given a sorted collection, perform binary search to find element x for which
    the following holds: item > key(x) and the value key(x) is the largest.
    Returns index of such an element.
    """
    lo = 0
    hi = len(sortedCollection)
    while lo < hi:
        mid = (lo + hi) // 2
        if item > key(sortedCollection[mid]):
            lo = mid + 1
        else:
            hi = mid
    return lo - 1

def buildShadows(lines: Iterable[AxialLine], boundaries: Iterable[AxialLine]) -> List[ShadowLine]:
    """
    Given an iterable of AxialLines, build their prolonged shadows. Shadows
    stop at the boundaries. Lines and boundaries are expected to be
    perpendicular to each other. This function assumes there is a boundary for
    every line.
    """
    boundaries = list(boundaries)
    boundaries.sort(key=lambda line: line.x)

    shadowLines: List[ShadowLine] = []
    for l in lines:
        # Extend to right
        righStart = lowerBound(boundaries, l.max, key=lambda line: line.x)
        for i in range(righStart, len(boundaries)):
            b = boundaries[i]
            if l.x in b and b.x > l.min:
                rightExtend = b.x
                break
        # Extend to left
        leftStart = upperBound(boundaries, l.min, key=lambda line: line.x)
        for i in range(leftStart, 0, -1):
            b  = boundaries[i]
            if l.x in b and b.x < l.max:
                leftExtend = b.x
                break
        assert rightExtend is not None
        assert leftExtend is not None
        shadowLines.append(ShadowLine(l, Interval(leftExtend, rightExtend)))
    return shadowLines

def trimShadows(shadows: Iterable[ShadowLine], boundaries: Iterable[AxialLine]) -> List[ShadowLine]:
    """
    Given an iterable of ShadowLines and Axial lines as boudaries, trim the
    shadows so they do not cross any boundary. Return new shadows.
    """
    boundaries = list(boundaries)
    boundaries.sort(key=lambda line: line.x)
    newShadows: List[ShadowLine] = []
    for l in shadows:
        # Trim right
        rightStart = upperBound(boundaries, l.line.min, key=lambda line: line.x)
        rightTrim = l.shadow.max
        for i in range(rightStart, len(boundaries)):
            b = boundaries[i]
            if l.line.x in b and l.shadow.min < b.x <= l.shadow.max:
                rightTrim = b.x
                break
        # Trim left
        leftStart = upperBound(boundaries, l.line.max, key=lambda line: line.x) - 1
        leftTrim = l.shadow.min
        for i in range(leftStart, 0, -1):
            b = boundaries[i]
            if l.line.x in b and l.shadow.min <= b.x < l.shadow.max:
                leftTrim = b.x
                break
        newShadows.append(ShadowLine(l.line, Interval(leftTrim, rightTrim)))
    return newShadows

class BoxPartitionLines:
    """
    Given a set of axially arranged non-overlapping boxes answers the query for
    the partition lines of each box.

    Partition line is a horizontal or a vertical line. The union of all
    partition lines splits the free space between the boxes such that the gaps
    between two neighboring boxes are split evenly. See the following
    illustration:

    +---+ | +----+ | +--------+
    |   | | |    | | |        |
    +---+ | |    | | |        |
    ------| |    | | +--------+
    +---+ | |    | |-----------
    |   | | |    |   |   +----+
    +---+ | +----+   |   +----+
    """

    def __init__(self, boxes: Dict[object, Box],
                 seedFilter: Callable[[object, object, bool, AxialLine], bool]=defaultSeedFilter,
                 safeHorizontalMargin: float=0, safeVerticalMargin: float=0) -> None:
        """
        Given a dictionary id -> box initializes the structure.

        Boxes are represented by a tuple (minx, miny, maxx, maxy)

        The margin guarantees there will be no partition line too close to edge
        (necessary to handle some pathological cases)
        """
        from kikit.common import shpBBoxExpand

        hstops, vstops = collectHardStops(boxes.values())
        hSafeStops, vSafeStops = collectHardStops([
            shpBBoxExpand(x, safeVerticalMargin, safeHorizontalMargin) for x in boxes.values()])
        hseeds, vseeds = collectSeedLines(boxes, seedFilter)
        hshadows = buildShadows(hseeds, chain(vstops, vSafeStops))
        vshadows = buildShadows(vseeds, chain(hstops, hSafeStops))

        hPartition = trimShadows(hshadows, chain(
            [x.shadowLine for x in vshadows], vSafeStops))
        vPartition = trimShadows(vshadows, chain(
            [x.shadowLine for x in hshadows], hSafeStops))

        self.query: Dict[object, Tuple[List[AxialLine], List[AxialLine]]] = { ident: ([], []) for ident in boxes.keys() }
        for l in hPartition:
            self.query[l.line.tag][0].append(
                AxialLine(l.line.x, l.shadow.min, l.shadow.max))
        for l in vPartition:
            self.query[l.line.tag][1].append(
                AxialLine(l.line.x, l.shadow.min, l.shadow.max))

    def partitionLines(self, ident: object) -> Tuple[List[AxialLine], List[AxialLine]]:
        """
        Return a tuple (horiz. lines, vert. lines) represented as AxialLine
        """
        return self.query[ident]

    @typing.no_type_check
    def _visualize(self, vars: Dict[str, Any]) -> None:
        """
        When debugging, you can invoke self._visualize(locals()) in __init__ to
        see what is happening.
        """
        import matplotlib.pyplot as plt

        for h in vars["hstops"]:
            plt.hlines(h.x, h.min, h.max, ["g"])
        for v in vars["vstops"]:
            plt.vlines(v.x, v.min, v.max, ["g"])

        plt.axis('equal')
        for h in vars["hshadows"]:
            plt.hlines(h.line.x, h.shadow.min, h.shadow.max, ["r"])
        for v in vars["vshadows"]:
            plt.vlines(v.line.x, v.shadow.min, v.shadow.max, ["r"])

        for h in vars["hseeds"]:
            plt.hlines(h.x, h.min, h.max)
        for v in vars["vseeds"]:
            plt.vlines(v.x, v.min, v.max)
        plt.show()



from typing import Tuple
from .common import KiLength
from .units import mm
from .sexpr import SExpr, findNode
from .defs import PAPER_DIMENSIONS

def getPageDimensionsFromAst(ast: SExpr) -> Tuple[KiLength, KiLength]:
    paperNode = findNode(ast, "paper")
    if paperNode is None:
        # KiCAD 5 board use "page" instead of "paper"
        paperNode = findNode(ast, "page")
    if paperNode is None:
        raise RuntimeError("Source document doesn't contain paper size information")
    value = paperNode.items[1].value
    if value == "User":
        size = (paperNode[2].value, paperNode[3].value)
        return tuple(int(float(x) * mm) for x in size)
    try:
        size = PAPER_DIMENSIONS[value]
        if len(paperNode.items) >= 3 and paperNode.items[2] == "portrait":
            size = (size[1], size[0])
        return tuple(int(x) for x in size)
    except KeyError:
        raise RuntimeError(f"Uknown paper size {value}") from None




import click
import os
import csv
import io
import glob
import traceback
from kikit.panelize_ui_sections import *

PKG_BASE = os.path.dirname(__file__)
PRESETS = os.path.join(PKG_BASE, "resources/panelizePresets")
IS_CLICK_V8 = click.__version__.startswith("8.")

# We would like to support both, click v7 and v8 in order to maximize
# compatibility. However, since click v8.1 there is breaking change in the way
# shell completion works. This functions hides the differences and should allow
# us to use both. Pass to it as **addCompatibleCompletion(completionFunction)
def addCompatibleShellCompletion(completionFn):
    if IS_CLICK_V8:
        import click.shell_completion
        def completion(*args, **kwargs):
            return [click.shell_completion.CompletionItem(x) for x in completionFn(*args, **kwargs)]
        return {"shell_complete": completionFn}
    else:
        return {"autocompletion": completionFn}

def splitStr(delimiter, escapeChar, s):
    """
    Splits s based on delimiter that can be escaped via escapeChar
    """
    # Let's use csv reader to implement this
    reader = csv.reader(io.StringIO(s), delimiter=delimiter, escapechar=escapeChar)
    # Unpack first line
    for x in reader:
        return x


class Section(click.ParamType):
    """
    A CLI argument type for overriding section parameters. Basically a semicolon
    separated list of `key: value` pairs. The first word might omit the key; in
    that case "type" key is used.
    """
    name = "parameter_list"

    def convert(self, value, param, ctx):
        if len(value.strip()) == 0:
            self.fail(f"{value} is not a valid argument specification",
                param, ctx)
        try:
            values = {}
            for i, pair in enumerate(splitStr(";", "\\", value)):
                if len(pair.strip()) == 0:
                    continue
                s = pair.split(":", 1)
                if i == 0 and len(s) == 1:
                    values["type"] = s[0].strip()
                    continue
                key, value = s[0].strip(), s[1].strip()
                values[key] = value
            return values
        except (TypeError, IndexError):
            self.fail(f"'{pair}' is not a valid key: value pair",
                param,
                ctx)

class HookPlugin(click.ParamType):
    """
    A CLI argument type for a HookPlugin. The format is <moduleName or
    path>:<plugin name>:<arg>. The arg is optional.
    """
    name = "<module>.<plugin>:[arg]"

    def convert(self, value, param, ctx):
        pieces = value.split(":", maxsplit=1)
        specPieces = pieces[0].rsplit(".", maxsplit=1)
        if len(specPieces) < 2:
            self.fail(f"{value} is not a valid plugin specification")
        module = specPieces[0]
        pluginName = specPieces[1]
        arg = "" if len(pieces) == 2 else pieces[1]
        return (module, pluginName, arg)


def completePath(prefix, fileSuffix=""):
    """
    This is rather hacky and  far from ideal, however, until Click 8 we probably
    cannot do much better.
    """
    paths = []
    for p in glob.glob(prefix + "*"):
        if os.path.isdir(p):
            paths.append(p + "/")
        elif p.endswith(fileSuffix):
            paths.append(p)
    return paths

def pathCompletion(fileSuffix=""):
    def f(ctx, args, incomplete):
        return completePath(incomplete, fileSuffix)
    return f

def completePreset(ctx, args, incomplete):
    presets = [":" + x.replace(".json", "")
        for x in os.listdir(PRESETS)
        if x.endswith(".json") and (x.startswith(incomplete) or x.startswith(incomplete[1:]))]
    if incomplete.startswith(":"):
        return presets
    return presets + completePath(incomplete, ".json")

def lastSectionPair(incomplete):
    """
    Given an incomplete command text of a section, return the last (possibly
    incomplete) key-value pair
    """
    lastSection = incomplete.split(";")[-1]
    x = [x.strip() for x in lastSection.split(":", 1)]
    if len(x) == 1:
        return x[0], ""
    return x

def hasNoSectionPair(incomplete):
    return ";" not in incomplete

def completeSection(section):
    def fun(ctx, args, incomplete):
        if incomplete.startswith("'"):
            incomplete = incomplete[1:]
        key, val = lastSectionPair(incomplete)

        candidates = []
        if hasNoSectionPair(incomplete):
            candidates.extend([x for x in section["type"].vals if x.startswith(incomplete)])
        if len(val) == 0:
            trimmedIncomplete = incomplete.rsplit(";", 1)[0]
            candidates.extend([trimmedIncomplete + x + ":"
                for x in section.keys() if x.startswith(key)])
        return candidates
    return fun

@click.command()
@click.argument("input", type=click.Path(dir_okay=False),
    **addCompatibleShellCompletion(pathCompletion(".kicad_pcb")))
@click.argument("output", type=click.Path(dir_okay=False),
    **addCompatibleShellCompletion(pathCompletion(".kicad_pcb")))
@click.option("--preset", "-p", multiple=True,
    help="A panelization preset file; use prefix ':' for built-in styles.",
    **addCompatibleShellCompletion(completePreset))
@click.option("--plugin", multiple=True, type=HookPlugin(),
    help="A hook plugin to use during the panelization",
    **addCompatibleShellCompletion(completePreset))
@click.option("--layout", "-l", type=Section(),
    help="Override layout settings.",
    **addCompatibleShellCompletion(completeSection(LAYOUT_SECTION)))
@click.option("--source", "-s", type=Section(),
    help="Override source settings.",
    **addCompatibleShellCompletion(completeSection(SOURCE_SECTION)))
@click.option("--tabs", "-t", type=Section(),
    help="Override tab settings.",
    **addCompatibleShellCompletion(completeSection(TABS_SECTION)))
@click.option("--cuts", "-c", type=Section(),
    help="Override cut settings.",
    **addCompatibleShellCompletion(completeSection(CUTS_SECTION)))
@click.option("--framing", "-r", type=Section(),
    help="Override framing settings.",
    **addCompatibleShellCompletion(completeSection(FRAMING_SECTION)))
@click.option("--tooling", "-o", type=Section(),
    help="Override tooling settings.",
    **addCompatibleShellCompletion(completeSection(TOOLING_SECTION)))
@click.option("--fiducials", "-f", type=Section(),
    help="Override fiducials settings.",
    **addCompatibleShellCompletion(completeSection(FIDUCIALS_SECTION)))
@click.option("--text", type=Section(),
    help="Override text settings.",
    **addCompatibleShellCompletion(completeSection(TEXT_SECTION)))
@click.option("--text2", type=Section(),
    help="Override text settings.",
    **addCompatibleShellCompletion(completeSection(TEXT_SECTION)))
@click.option("--text3", type=Section(),
    help="Override text settings.",
    **addCompatibleShellCompletion(completeSection(TEXT_SECTION)))
@click.option("--text4", type=Section(),
    help="Override text settings.",
    **addCompatibleShellCompletion(completeSection(TEXT_SECTION)))
@click.option("--copperfill", "-u", type=Section(),
    help="Override copper fill settings.",
    **addCompatibleShellCompletion(completeSection(COPPERFILL_SECTION)))
@click.option("--page", "-P", type=Section(),
    help="Override page settings.",
    **addCompatibleShellCompletion(completeSection(POST_SECTION)))
@click.option("--post", "-z", type=Section(),
    help="Override post processing settings.",
    **addCompatibleShellCompletion(completeSection(POST_SECTION)))
@click.option("--debug", type=Section(),
    help="Include debug traces or drawings in the panel.",
    **addCompatibleShellCompletion(completeSection(DEBUG_SECTION)))
@click.option("--dump", "-d", type=click.Path(file_okay=True, dir_okay=False),
    help="Dump constructured preset into a JSON file.")
def panelize(input, output, preset, plugin, layout, source, tabs, cuts, framing,
             tooling, fiducials, text, text2, text3, text4, copperfill, page,
             post, debug, dump):
    """
    Panelize boards
    """
    try:
        # Hide the import in the function to make KiKit start faster
        from kikit import panelize_ui_impl as ki
        import sys
        from kikit.common import fakeKiCADGui
        app = fakeKiCADGui()

        preset = ki.obtainPreset(preset,
            layout=layout, source=source, tabs=tabs, cuts=cuts, framing=framing,
            tooling=tooling, fiducials=fiducials, text=text,text2=text2,
            text3=text3, text4=text4, copperfill=copperfill, page=page,
            post=post, debug=debug)

        doPanelization(input, output, preset, plugin)

        if (dump):
            with open(dump, "w", encoding="utf-8") as f:
                f.write(ki.dumpPreset(preset))
    except Exception as e:
        import sys
        from kikit.panelize import NonFatalErrors
        if isinstance(e, NonFatalErrors):
            sys.stderr.write(str(e) + "\n")
        else:
            sys.stderr.write("An error occurred: " + str(e) + "\n")
            sys.stderr.write("No output files produced\n")
        if isinstance(preset, dict) and preset["debug"]["trace"]:
            traceback.print_exc(file=sys.stderr)
        sys.exit(1)

def doPanelization(input, output, preset, plugins=[]):
    """
    The panelization logic is separated into a separate function so we can
    handle errors based on the context; e.g., CLI vs GUI
    """
    from kikit import panelize_ui_impl as ki
    from kikit.panelize import Panel, NonFatalErrors, PanelError
    import pcbnew
    from pcbnew import LoadBoard
    from itertools import chain

    if preset["debug"]["drawtabfail"]:
        import kikit.substrate
        kikit.substrate.TABFAIL_VISUAL = True

    board = LoadBoard(input)
    if preset["debug"]["deterministic"]:
        pcbnew.KIID.SeedGenerator(42)
    if board is None:
        raise PanelError(f"Cannot load board {input}. Check if the path is correct or if you have permissions to read it.")
    panel = Panel(output)

    useHookPlugins = ki.loadHookPlugins(plugins, board, preset)

    useHookPlugins(lambda x: x.prePanelSetup(panel))

    # Register extra footprints for annotations
    for tabFootprint in preset["tabs"]["tabfootprints"]:
        panel.annotationReader.registerTab(tabFootprint.lib, tabFootprint.footprint)

    panel.inheritDesignSettings(board)
    panel.inheritProperties(board)
    panel.inheritTitleBlock(board)
    panel.inheritLayerNames(board)

    useHookPlugins(lambda x: x.afterPanelSetup(panel))

    sourceArea = ki.readSourceArea(preset["source"], board)
    substrates, framingSubstrates, backboneCuts = \
        ki.buildLayout(preset, panel, input, sourceArea)

    useHookPlugins(lambda x: x.afterLayout(panel, substrates))

    tabCuts = ki.buildTabs(preset, panel, substrates, framingSubstrates)

    useHookPlugins(lambda x: x.afterTabs(panel, tabCuts, backboneCuts))

    preFrameSubstrate = panel.boardSubstrate.substrates

    frameCuts = ki.buildFraming(preset, panel)

    useHookPlugins(lambda x: x.afterFraming(panel, frameCuts))

    ki.buildTabFillets(preset, panel, preFrameSubstrate)

    ki.buildTooling(preset, panel)
    ki.buildFiducials(preset, panel)
    for textSection in ["text", "text2", "text3", "text4"]:
        ki.buildText(preset[textSection], panel)
    ki.buildPostprocessing(preset["post"], panel)

    ki.makeTabCuts(preset, panel, tabCuts)
    ki.makeOtherCuts(preset, panel, chain(backboneCuts, frameCuts))

    useHookPlugins(lambda x: x.afterCuts(panel))

    ki.buildCopperfill(preset["copperfill"], panel)

    ki.setStackup(preset["source"], panel)
    ki.setPageSize(preset["page"], panel, board)
    ki.positionPanel(preset["page"], panel)

    ki.runUserScript(preset["post"], panel)
    useHookPlugins(lambda x: x.finish(panel))

    ki.buildDebugAnnotation(preset["debug"], panel)

    panel.save(reconstructArcs=preset["post"]["reconstructarcs"],
               refillAllZones=preset["post"]["refillzones"],
               edgeWidth=preset["post"]["edgewidth"])

    if panel.hasErrors():
        raise NonFatalErrors(panel.errors)


@click.command()
@click.argument("input", type=click.Path(dir_okay=False))
@click.argument("output", type=click.Path(dir_okay=False))
@click.option("--source", "-s", type=Section(),
    help="Specify source settings.")
@click.option("--page", "-P", type=Section(),
    help="Override page settings.",
    **addCompatibleShellCompletion(completeSection(POST_SECTION)))
@click.option("--debug", type=Section(),
    help="Include debug traces or drawings in the panel.")
@click.option("--keepAnnotations/--stripAnnotations", default=True,
    help="Do not strip annotations" )
@click.option("--preserveArcs/--looseArcs", default=True,
    help="Preserve arcs in the files" )
def separate(input, output, source, page, debug, keepannotations, preservearcs):
    """
    Separate a single board out of a multi-board design. The separated board is
    placed in the middle of the sheet.

    You can specify the board via bounding box or annotation. See documentation
    for further details on usage.
    """
    try:
        from kikit import panelize_ui_impl as ki
        from kikit.panelize import Panel, NonFatalErrors
        from kikit.units import mm
        import pcbnew
        from pcbnew import LoadBoard, VECTOR2I
        from kikit.common import fakeKiCADGui
        app = fakeKiCADGui()

        preset = ki.obtainPreset([], validate=False, source=source, page=page, debug=debug)

        board = LoadBoard(input)
        if preset["debug"]["deterministic"]:
            pcbnew.KIID.SeedGenerator(42)
        sourceArea = ki.readSourceArea(preset["source"], board)

        panel = Panel(output)
        panel.inheritDesignSettings(board)
        panel.inheritProperties(board)
        panel.inheritTitleBlock(board)
        panel.inheritLayerNames(board)

        destination = VECTOR2I(150 * mm, 100 * mm)
        panel.appendBoard(input, destination, sourceArea,
            interpretAnnotations=(not keepannotations),
            netRenamer=lambda i, x: x,
            refRenamer=lambda i, x: x)
        ki.setStackup(preset["source"], panel)
        ki.setPageSize(preset["page"], panel, board)
        ki.positionPanel(preset["page"], panel)

        panel.save(reconstructArcs=preservearcs)

        if panel.hasErrors():
            raise NonFatalErrors(panel.errors)
    except Exception as e:
        import sys
        sys.stderr.write("An error occurred: " + str(e) + "\n")
        sys.stderr.write("No output files produced\n")
        if isinstance(preset, dict) and preset["debug"]["trace"]:
            traceback.print_exc(file=sys.stderr)
        sys.exit(1)



from __future__ import annotations
from typing import Dict, List, Set, Union
import numpy as np
import pcbnew
from kikit import units # type: ignore

from kikit.common import fromMm, readParameterList

KString = Union[str, pcbnew.UTF8]

def readKiKitProps(footprint):
    """
    Given a footprint, returns a string containing KiKit annotations.
    Annotations are in fields starting with prefix `KIKIT:`.

    Returns a dictionary of key-value pairs.
    """
    for x in footprint.GraphicalItems():
        if not isinstance(x, pcbnew.PCB_FIELD) and not isinstance(x, pcbnew.PCB_TEXT):
            continue
        text = x.GetText()
        if text.startswith("KIKIT:"):
            return readParameterList(text[len("KIKIT:"):])
    return {}

class KiKitAnnotation:
    pass

class TabAnnotation(KiKitAnnotation):
    def __init__(self, ref, origin, direction, width, maxLength=fromMm(100), props={}):
        self.ref = ref
        self.origin = origin
        self.direction = direction
        self.width = width
        self.maxLength = maxLength
        self.props = props

    @staticmethod
    def fromFootprint(footprint):
        origin = footprint.GetPosition()
        radOrientation = footprint.GetOrientation().AsRadians()
        direction = (np.cos(radOrientation), -np.sin(radOrientation))
        props = readKiKitProps(footprint)
        if "width" not in props:
            raise ValueError("Tab annotation must a KiKit annotation text with 'width' property defined")
        width = units.readLength(props["width"])
        return TabAnnotation(footprint.GetReference(), origin, direction, width, props=props)

class AnnotationReader:
    """
    Instance of this class can convert footprints into KiCAD annotations. Having
    a class instead of a function allows the users to specify source footprints.
    """
    def __init__(self) -> None:
        self._tabSources: Dict[str, Set[str]] = {}
        self._boardSources: Dict[str, Set[str]] = {}

    def registerTab(self, lib: KString, footprint: KString) -> None:
        self._register(self._tabSources, lib, footprint)

    def registerBoard(self, lib: KString, footprint: KString) -> None:
        self._register(self._boardSources, lib, footprint)

    def _register(self, where: Dict[str, Set[str]], lib: KString, footprint: KString) -> None:
        libstr = str(lib).lower()
        footprintstr = str(footprint).lower()

        if libstr not in where:
            where[libstr] = set()
        where[libstr].add(footprintstr)

    def _isIn(self, where: Dict[str, Set[str]], lib: str, footprint: str) -> bool:
        lib = lib.lower()
        footprint = footprint.lower()
        if lib not in where:
            return False
        return footprint in where[lib]


    def isAnnotation(self, footprint: pcbnew.FOOTPRINT) -> bool:
        """
        Given a footprint, decide if it is KiKit annotation
        """
        info = footprint.GetFPID()
        lib = str(info.GetLibNickname())
        fname = str(info.GetLibItemName())

        for aType in [self._tabSources, self._boardSources]:
            if self._isIn(aType, lib, fname):
                return True
        return False

    def convertToAnnotation(self, footprint: pcbnew.FOOTPRINT) -> List[KiKitAnnotation]:
        """
        Given a footprint, convert it into an annotation. One footprint might
        represent zero or multiple annotations, so this function returns a list.
        """
        info = footprint.GetFPID()
        lib = str(info.GetLibNickname())
        fname = str(info.GetLibItemName())

        if self._isIn(self._tabSources, lib, fname):
            return [TabAnnotation.fromFootprint(footprint)]
        # We ignore Board annotation
        return []


    @staticmethod
    def getDefault() -> AnnotationReader:
        """
        Return AnnotationReader object initialized with default annotations registered
        """
        r = AnnotationReader()
        r.registerTab("kikit", "Tab")
        r.registerBoard("kikit", "Board")

        # Since KiCAD 7.0 the plugins installed via the Plugin and Content Manager get prepended with `PCM_`
        r.registerTab("PCM_kikit", "Tab")
        r.registerBoard("PCM_kikit", "Board")
        return r




from __future__ import annotations

import json
import os
import re
import shutil
import subprocess
import sys
import tempfile
from dataclasses import dataclass, field
from typing import Any, Dict, Iterable, List, Optional, TextIO, Tuple
from pathlib import Path

import pcbnew
from kikit.pcbnew_utils import EDA_UNITS_MM, EDA_UNITS_INCH, getItemDescription, resolveItem

from kikit.common import fromMm, toMm
from kikit.drc_ui import ReportLevel

NULL_UUID = "00000000-0000-0000-0000-000000000000"

def _find_kicad_cli() -> Optional[str]:
    """Locate the kicad-cli binary."""
    # Try PATH first
    result = shutil.which("kicad-cli")
    if result is not None:
        return result

    # Derive from the _pcbnew native module location
    try:
        import importlib.util
        spec = importlib.util.find_spec("_pcbnew")
        if spec is None or spec.origin is None:
            return None
        bindir = Path(os.path.realpath(spec.origin)).parent

        # macOS: _pcbnew.kiface is in Contents/Frameworks/, kicad-cli is in
        # Contents/MacOS/
        if "Contents/Frameworks" in str(bindir):
            bindir = bindir.parent / "MacOS"

        for name in ("kicad-cli", "kicad-cli.exe"):
            candidate = bindir / name
            if candidate.is_file():
                return str(candidate)
    except Exception:
        pass

    return None

ItemFingerprint = Tuple[int, int, str]

def roundCoord(x: int) -> int:
    # KiCAD doesn't round the values, it just cuts the decimal places. So let's
    # emulate that
    return round(x - 50, -4)


def getItemFingerprint(item: pcbnew.BOARD_ITEM):
    return (roundCoord(item.GetPosition()[0]), # Round down, since the output does the same
            roundCoord(item.GetPosition()[1]), # Round down, since the output does the same
            getItemDescription(item))

def collectFingerprints(board: pcbnew.BOARD) -> Dict[ItemFingerprint, pcbnew.BOARD_ITEM]:
    """
    Traverse the board and collect fingerprints of all items in the board
    """
    fingerprints = {}
    def collect(items: Iterable[pcbnew.BOARD_ITEM]) -> None:
        nonlocal fingerprints
        for x in items:
            fingerprints[getItemFingerprint(x)] = x

    collect(board.GetDrawings())
    collect(board.GetFootprints())
    for f in board.GetFootprints():
        collect(f.Pads())
        collect(f.GraphicalItems())
        collect(f.Zones())
        collect([f.Reference(), f.Value()])
    collect(board.GetTracks())
    collect(board.Zones())

    return fingerprints

@dataclass
class DrcExclusion:
    type: str
    position: pcbnew.VECTOR2I
    objects: List[pcbnew.BOARD_ITEM] = field(default_factory=list)

    def eqRepr(self):
        if len(self.objects) == 0:
            return (self.type, ())
        if len(self.objects) == 1:
            objRepr = str(self.objects[0].m_Uuid.AsString()) if isinstance(self.objects[0], pcbnew.BOARD_ITEM) else self.objects[0]
            return (self.type, objRepr)
        if len(self.objects) == 2 or self.type in ["starved_thermal"]:
            return (self.type, frozenset(str(x.m_Uuid.AsString()) for x in self.objects))
        raise RuntimeError("Unsupported exclusion object count")

@dataclass
class Violation:
    type: str
    description: str
    rule: str
    severity: str
    objects: List[pcbnew.BOARD_ITEM] = field(default_factory=list)

    def format(self, units: Any) -> str:
        head = f"[{self.type}]: {self.description}\n    {self.rule}; Severity: {self.severity}"
        tail = "\n".join(["    " + self._formatObject(x, units) for x in self.objects])
        return "\n".join([head] + [tail])

    def _formatObject(self, obj: pcbnew.BOARD_ITEM, units: Any) -> str:
        p = obj.GetPosition()
        pos = "unknown"
        if units == EDA_UNITS_MM:
            pos = f"{toMm(p[0]):.4f} mm, {toMm(p[1]):.4f} mm"
        if units == EDA_UNITS_INCH:
            pos = f"{pcbnew.ToMils(p[0]):.1f} mil, {pcbnew.ToMils(p[1]):.1f} mil"
        return f"@({pos}): {getItemDescription(obj, units)}"

    def eqRepr(self):
        if len(self.objects) == 0: # E.g., copper sliver has no related objects
            return (self.type, self.description)
        if len(self.objects) == 1:
            return (self.type, self.objects[0].m_Uuid.AsString())
        if len(self.objects) == 2:
            return (self.type, frozenset(str(x.m_Uuid.AsString()) for x in self.objects))
        raise RuntimeError("Unsupported violation object count")

@dataclass
class DrcReport:
    """
    Lists of DRC type violations
    """
    drc: List[Violation]
    unconnected: List[Violation]
    footprint: List[Violation]

    def items(self) -> Iterable[Tuple[str, List[Violation]]]:
        return {
            "drc": self.drc,
            "unconnected": self.unconnected,
            "footprint": self.footprint
        }.items()

    def pruneExclusions(self, exclusions: List[DrcExclusion]) -> None:
        """
        Given a list of exclusions, prune the report.
        """
        prints = set(x.eqRepr() for x in exclusions)
        self.drc = [x for x in self.drc if x.eqRepr() not in prints]
        self.unconnected = [x for x in self.unconnected if x.eqRepr() not in prints]
        self.footprint = [x for x in self.footprint if x.eqRepr() not in prints]

@dataclass
class CliViolation:
    """A DRC violation parsed from kicad-cli JSON output."""
    type: str
    description: str
    severity: str
    items: List[Dict] = field(default_factory=list)

    def format(self, units: Any) -> str:
        head = f"[{self.type}]: {self.description}\n    Severity: {self.severity}"
        lines = []
        for item in self.items:
            pos = item.get("pos", {})
            x, y = pos.get("x", 0), pos.get("y", 0)
            if units == EDA_UNITS_MM:
                pos_str = f"{x:.4f} mm, {y:.4f} mm"
            elif units == EDA_UNITS_INCH:
                pos_str = f"{x / 0.0254:.1f} mil, {y / 0.0254:.1f} mil"
            else:
                pos_str = "unknown"
            lines.append(f"    @({pos_str}): {item.get('description', '')}")
        return "\n".join([head] + lines)

    def eqRepr(self):
        uuids = [item["uuid"] for item in self.items if "uuid" in item]
        if len(uuids) == 0:
            return (self.type, self.description)
        if len(uuids) == 1:
            return (self.type, uuids[0])
        return (self.type, frozenset(uuids))

@dataclass
class CliDrcExclusion:
    """A DRC exclusion parsed from the project file without pcbnew."""
    type: str
    uuids: List[str] = field(default_factory=list)

    def eqRepr(self):
        if len(self.uuids) == 0:
            return (self.type, ())
        if len(self.uuids) == 1:
            return (self.type, self.uuids[0])
        return (self.type, frozenset(self.uuids))

def _readExclusionsFromProjectFile(boardFilename: str) -> List[CliDrcExclusion]:
    """Read DRC exclusions from .kicad_pro without needing pcbnew."""
    projectFilename = os.path.splitext(boardFilename)[0] + '.kicad_pro'
    try:
        with open(projectFilename, encoding="utf-8") as f:
            project = json.load(f)
    except FileNotFoundError:
        raise FileNotFoundError(
            f"Board '{boardFilename}' has no project, cannot read DRC exclusions")
    try:
        exclusions = project["board"]["design_settings"]["drc_exclusions"]
    except KeyError:
        return []
    if len(exclusions) > 0 and isinstance(exclusions[0], list):
        exclusions = [x[0] for x in exclusions]
    result = []
    for e in exclusions:
        items = e.split("|")
        uuids = [u for u in items[3:] if u != NULL_UUID]
        result.append(CliDrcExclusion(items[0], uuids))
    return result

def _runCliDrc(kicadCli: str, boardFile: str, strict: bool) -> DrcReport:
    """Run DRC via kicad-cli and return a DrcReport with CliViolation objects."""
    with tempfile.NamedTemporaryFile(mode="w", suffix=".json", delete=False) as tmp:
        tmpName = tmp.name
    try:
        cmd = [kicadCli, "pcb", "drc",
               "--format", "json",
               "--output", tmpName,
               "--units", "mm",
               "--severity-all",
               "--exit-code-violations"]
        if strict:
            cmd.append("--all-track-errors")
        cmd.append(boardFile)

        proc = subprocess.run(cmd, capture_output=True, text=True)
        # Exit 0 = clean, 5 = violations found (both ok), anything else = error
        if proc.returncode not in (0, 5):
            raise RuntimeError(
                f"kicad-cli DRC failed (exit {proc.returncode}): "
                f"{proc.stderr.strip() or proc.stdout.strip()}")

        with open(tmpName, encoding="utf-8") as f:
            data = json.load(f)
    finally:
        try:
            os.unlink(tmpName)
        except OSError:
            pass

    def parseViolations(items: List[Dict]) -> List[CliViolation]:
        return [CliViolation(
            type=v["type"],
            description=v["description"],
            severity=v["severity"],
            items=v.get("items", [])
        ) for v in items]

    return DrcReport(
        drc=parseViolations(data.get("violations", [])),
        unconnected=parseViolations(data.get("unconnected_items", [])),
        footprint=parseViolations(data.get("schematic_parity", []))
    )

def readBoardItem(text: str,
                  fingerprints: Dict[ItemFingerprint, pcbnew.BOARD_ITEM]) \
                    -> pcbnew.BOARD_ITEM:
    """
    Given DRC report object description, try to find it in the board
    """
    itemMatch = re.match(r'\s*@\((-?\d*(\.\d*)?) mm, (-?\d*(\.\d*)?) mm\): (.*)$', text)
    if itemMatch is None:
        raise RuntimeError(f"Cannot parse board item from '{text}'")
    posX = float(itemMatch.group(1))
    posY = float(itemMatch.group(3))
    descr = itemMatch.group(5)
    fPrint = (roundCoord(fromMm(posX)), roundCoord(fromMm(posY)), str(descr))
    try:
        return fingerprints[fPrint]
    except KeyError:
        raise RuntimeError(f"Cannot find board item from '{text}', fingerprint: '{fPrint}'") # from None

def readViolations(reportFile: TextIO,
                   fingerprints: Dict[ItemFingerprint, pcbnew.BOARD_ITEM]) \
                        -> Tuple[str, List[Violation]]:
    violations = []
    line = reportFile.readline()
    while True:
        headerMatch = re.match(r'\[(.*)\]: (.*)\n', line)
        if headerMatch is None:
            break
        line = reportFile.readline()
        bodyMatch = re.match(r'\s*(.*); (Severity: )?(.*)', line)
        if bodyMatch is None:
            break
        v = Violation(
            type = headerMatch.group(1),
            description = headerMatch.group(2),
            rule = bodyMatch.group(1),
            severity = bodyMatch.group(3))
        line = reportFile.readline()
        while line.startswith("    "):
            v.objects.append(readBoardItem(line, fingerprints))
            line = reportFile.readline()
        violations.append(v)

    return line, violations

def readReport(reportFile: TextIO, board: pcbnew.BOARD) -> DrcReport:
    fingerprints = collectFingerprints(board)
    drcV: List[Violation] = []
    unconnectedV: List[Violation] = []
    footprintV: List[Violation] = []
    line = reportFile.readline()
    while True:
        if len(line) == 0:
            break
        if re.match(r'\*\* Found \d+ DRC violations \*\*', line):
            line, drcV = readViolations(reportFile, fingerprints)
            continue
        if re.match(r'\*\* Found \d+ unconnected pads \*\*', line):
            line, unconnectedV = readViolations(reportFile, fingerprints)
            continue
        if re.match(r'\*\* Found \d+ Footprint errors \*\*', line):
            line, footprintV = readViolations(reportFile, fingerprints)
        line = reportFile.readline()
    return DrcReport(drcV, unconnectedV, footprintV)

def runBoardDrc(board: pcbnew.BOARD, strict: bool) -> DrcReport:
    projectPath = Path(board.GetFileName()).resolve().with_suffix(".kicad_pro")
    pcbnew.GetSettingsManager().LoadProject(str(projectPath))
    with tempfile.NamedTemporaryFile(mode="w+", delete=False) as tmpFile:
        try:
            tmpFile.close()
            result = pcbnew.WriteDRCReport(board, tmpFile.name,
                                           EDA_UNITS_MM, strict)
            if not result:
                raise RuntimeError("Cannot run DRC: Unspecified KiCAD error")
            with open(tmpFile.name, encoding="utf-8") as f:
                report = readReport(f, board)
        finally:
            tmpFile.close()
            os.unlink(tmpFile.name)
    return report

def deserializeExclusion(exclusionText: str, board: pcbnew.BOARD) -> DrcExclusion:
    items = exclusionText.split("|")
    objects = [resolveItem(board, pcbnew.KIID(x)) for x in items[3:]]
    objects = [x for x in objects if x is not None]
    return DrcExclusion(items[0],
                        pcbnew.VECTOR2I(int(items[1]), int(items[2])),
                        objects)

def serializeExclusion(exclusion: DrcExclusion) -> str:
    objIds = [x.m_Uuid.AsString() for x in exclusion.objects]
    while len(objIds) < 2:
        objIds.append("00000000-0000-0000-0000-000000000000")
    return "|".join([
        str(exclusion.type),
        str(exclusion.position[0]),
        str(exclusion.position[1])] + objIds
    )

def readBoardDrcExclusions(board: pcbnew.BOARD) -> List[DrcExclusion]:
    projectFilename = os.path.splitext(board.GetFileName())[0]+'.kicad_pro'
    try:
        with open(projectFilename, encoding="utf-8") as f:
            project = json.load(f)
    except FileNotFoundError:
        raise FileNotFoundError(f"Board '{board.GetFileName()}' has no project, cannot read DRC exclusions")
    try:
        exclusions = project["board"]["design_settings"]["drc_exclusions"]
    except KeyError:
        return [] # There are no exclusions
    if len(exclusions) > 0 and isinstance(exclusions[0], list):
        exclusions = [x[0] for x in exclusions]
    return [deserializeExclusion(e, board) for e in exclusions]

def runImpl(board, useMm, ignoreExcluded, strict, level, yieldViolation):
    import faulthandler
    faulthandler.enable(sys.stderr)

    units = EDA_UNITS_MM if useMm else EDA_UNITS_INCH
    boardFile = board.GetFileName()

    kicadCli = _find_kicad_cli()
    if kicadCli is not None:
        report = _runCliDrc(kicadCli, boardFile, strict)
        if ignoreExcluded:
            report.pruneExclusions(_readExclusionsFromProjectFile(boardFile))
    else:
        report = runBoardDrc(board, strict)
        if ignoreExcluded:
            report.pruneExclusions(readBoardDrcExclusions(board))

    failed = False
    errorName = {
        "drc": "DRC violations",
        "unconnected": "unconnected pads",
        "footprint": "footprints errors"
    }
    for k, v in report.items():
        if len(v) == 0:
            continue
        failed = False
        failedCases = []
        for x in v:
            thisFailed = False
            if level == ReportLevel.warning and x.severity == "warning":
                thisFailed = True
            if x.severity == "error":
                thisFailed = True
            if thisFailed:
                failedCases.append(x)
            failed = failed or thisFailed
        if failedCases:

            msg = f"** Found {len(failedCases)} {errorName[k]}: **\n"
            msg += "\n".join([x.format(units) for x in failedCases])
            yieldViolation(msg)
    return failed



# Simple, rather hacky parser for eeshema files (.sch). This is a mid-term
# workaround before proper support for scripting is introduced into Eeschema

import shlex
import os

class EeschemaException(Exception):
    pass

def getField(component, field):
    for f in component["fields"]:
        n = f["number"]
        if field == "Reference" and n == 0:
            return f["text"]
        if field == "Value" and n == 1:
            return f["text"]
        if field == "Footprint" and n == 2:
            return f["text"]
        if field == "Datasheet" and n == 3:
            return f["text"]
        if "name" in f and field == f["name"]:
            return f["text"]
    return None

def readEeschemaLine(file):
    line = ""
    quotationOpen = False
    while True:
        c = file.read(1)
        if c is None:
            raise EeschemaException("Cannot parse EEschema, line expected, got EOF")
        if c == '"':
            quotationOpen = not quotationOpen
        if c == '\n' and not quotationOpen:
            break
        line += c
    return line.strip()

def readHeader(file):
    VERSION_STRING = "EESchema Schematic File Version"
    DESCR_STRING = "$Descr "
    header = {}
    while True:
        line = readEeschemaLine(file)
        if line == "$EndDescr":
            return header

        if line.startswith(VERSION_STRING):
            header["version"] = line[len(VERSION_STRING):].strip()
        elif line.startswith("LIBS:"):
            header["libs"] = header.get("libs", []) + [line[len("LIBS:"):].strip()]
        elif line.startswith("EELAYER"):
            pass
        elif line.startswith(DESCR_STRING):
            header["size"] = line[len(DESCR_STRING):].split()
        elif line.startswith("Sheet"):
            items = shlex.split(line)
            header["sheet"] = (int(items[1]), int(items[2]))
        elif line.startswith("Title"):
            header["title"] = shlex.split(line)[1]
        elif line.startswith("Date"):
            header["date"] = shlex.split(line)[1]
        elif line.startswith("Comp"):
            header["company"] = shlex.split(line)[1]
        elif line.startswith("Rev"):
            header["revision"] = shlex.split(line)[1]
        elif line.startswith("Comment1"):
            header["comment1"] = shlex.split(line)[1]
        elif line.startswith("Comment2"):
            header["comment2"] = shlex.split(line)[1]
        elif line.startswith("Comment3"):
            header["comment3"] = shlex.split(line)[1]
        elif line.startswith("Comment4"):
            header["comment4"] = shlex.split(line)[1]
        elif line.startswith("encoding"):
            header["encoding"] = shlex.split(line)[1]
        else:
            raise EeschemaException(f"Unexpected line: '{line}'")

def readComponent(file, sheetPath=""):
    component = {}
    while True:
        line = readEeschemaLine(file)
        if line == "$EndComp":
            return component

        if line.startswith("L"):
            items = shlex.split(line)
            component["reference"] = items[2]
            component["name"] = items[1]
        elif line.startswith("U"):
            items = shlex.split(line)
            component["u"] = items[3]
            component["unit"] = int(items[1])
        elif line.startswith("P"):
            items = shlex.split(line)
            component["position"] = (int(items[1]), int(items[2]))
        elif line.startswith("F"):
            items = shlex.split(line)
            field = {
                "number": int(items[1]),
                "text": items[2],
                "orientation": items[3],
                "position": (int(items[4]), int(items[5])),
                "size": items[6],
                "flags": items[7],
                "justify": items[8],
                "style": items[9]
            }
            if field["number"] >= 4:
                field["name"] = items[10]
            component["fields"] = component.get("fields", []) + [field]
        elif line.startswith("AR"):
            # Hierarchical sheet reference. We assume all sheets are used within
            # the project, therefore we do not check validity of path
            items = shlex.split(line)
            path = None
            ref = None
            for item in items:
                if item.startswith('Path='):
                    path = item[len('Path='):]
                    break
            for item in items:
                if item.startswith('Ref='):
                    ref = item[len('Ref='):]
                    break
            if path is not None and ref is not None:
                compPath = sheetPath + "/" + component["u"]
                if path == compPath:
                    component["reference"] = ref
        else:
            items = shlex.split(line)
            try:
                int(items[0])
                if len(items) == 3 and items[0] == "1":
                    # Duplicate position entry
                    pass
                if len(items) == 4:
                    component["orientation"] = [int(x) for x in items[1:]]
            except ValueError:
                raise EeschemaException(f"Unexpected line: '{line}'")

def readSheet(file):
    # Note that the parser is incomplete an only extracts filename and reference
    # from the sheet as it is all we need currently
    sheet = {}
    while True:
        line = readEeschemaLine(file)
        if line == "$EndSheet":
            return sheet
        if line.startswith("F1 "):
            items = shlex.split(line)
            sheet["f1"] = items[1]
        elif line.startswith("U "):
            sheet["u"] = shlex.split(line)[1]

def extractComponents(filename, path=""):
    """
    Extract all components from the schematics
    """
    components = []
    sheets = []
    with open(filename, "r", encoding="utf-8") as file:
        header = readHeader(file)
        while True:
            line = file.readline()
            if not line:
                break
            line = line.strip()
            if line.startswith("$Comp"):
                components.append(readComponent(file, path))
            if line.startswith("$Sheet"):
                sheets.append(readSheet(file))
    for s in sheets:
        dirname = os.path.dirname(filename)
        sheetfilename = os.path.join(dirname, s["f1"])
        components += extractComponents(sheetfilename, path + "/" + s["u"])
    return components

def getUnit(component):
    return component["unit"]

def getReference(component):
    return component["reference"]



# Based on https://github.com/KiCad/kicad-source-mirror/blob/master/demos/python_scripts_examples/gen_gerber_and_drill_files_board.py
import sys
import os
import pcbnew
from pcbnew import *
from kikit.pcbnew_utils import DXF_UNITS_MM

from kikit.defs import Layer

fullGerberPlotPlan = [
    # name, id, comment
    ("CuTop", Layer.F_Cu, "Top layer"),
    ("CuBottom", Layer.B_Cu, "Bottom layer"),
    ("PasteBottom", Layer.B_Paste, "Paste Bottom"),
    ("PasteTop", Layer.F_Paste, "Paste top"),
    ("SilkTop", Layer.F_SilkS, "Silk top"),
    ("SilkBottom", Layer.B_SilkS, "Silk top"),
    ("MaskBottom", Layer.B_Mask, "Mask bottom"),
    ("MaskTop", Layer.F_Mask, "Mask top"),
    ("EdgeCuts", Layer.Edge_Cuts, "Edges"),
    ("CmtUser", Layer.Cmts_User, "V-CUT")
]

exportSettingsJlcpcb = {
    "UseGerberProtelExtensions": True,
    "UseAuxOrigin": True,
    "ExcludeEdgeLayer": True,
    "MinimalHeader": False,
    "NoSuffix": False,
    "MergeNPTH": False,
    "MapFileFormat": PLOT_FORMAT_GERBER,
    "ZerosFormat": GENDRILL_WRITER_BASE.DECIMAL_FORMAT,
    "SubstractMaskFromSilk": True
}

exportSettingsPcbway = {
    "UseGerberProtelExtensions": True,
    "UseAuxOrigin": False,
    "ExcludeEdgeLayer": True,
    "MinimalHeader": True,
    "NoSuffix": True,
    "MergeNPTH": False,
    "MapFileFormat": PLOT_FORMAT_PDF,
    "ZerosFormat": GENDRILL_WRITER_BASE.SUPPRESS_LEADING,
}

exportSettingsOSHPark = {
    "UseGerberProtelExtensions": True,
    "UseAuxOrigin": False,
    "ExcludeEdgeLayer": True,
    "MinimalHeader": False,
    "NoSuffix": False,
    "MergeNPTH": True,
    "MapFileFormat": PLOT_FORMAT_PDF,
    "ZerosFormat": GENDRILL_WRITER_BASE.DECIMAL_FORMAT,
}


def hasCopper(plotPlan):
    for _, layer, _ in plotPlan:
        if layer in [Layer.F_Cu, Layer.B_Cu]:
            return True
    return False

def setExcludeEdgeLayer(plotOptions, excludeEdge):
    try:
        plotOptions.SetExcludeEdgeLayer(excludeEdge)
    except AttributeError:
        if excludeEdge:
            plotOptions.SetLayerSelection(LSET())
        else:
            plotOptions.SetLayerSelection(LSET(Layer.Edge_Cuts))

def gerberImpl(boardfile, outputdir, plot_plan=fullGerberPlotPlan, drilling=True, settings=exportSettingsJlcpcb):
    """
    Export board to gerbers.

    If no output dir is specified, use '<board file>-gerber'
    """
    basename = os.path.basename(boardfile)
    if outputdir:
        plotDir = outputdir
    else:
        plotDir = basename + "-gerber"
    plotDir = os.path.abspath(plotDir)

    board = LoadBoard(boardfile)

    pctl = PLOT_CONTROLLER(board)
    popt = pctl.GetPlotOptions()

    popt.SetOutputDirectory(plotDir)

    popt.SetPlotFrameRef(False)
    popt.SetSketchPadLineWidth(FromMM(0.35))
    popt.SetAutoScale(False)
    popt.SetScale(1)
    popt.SetMirror(False)
    popt.SetUseGerberAttributes(False)
    popt.SetIncludeGerberNetlistInfo(True)
    popt.SetCreateGerberJobFile(True)
    popt.SetUseGerberProtelExtensions(settings["UseGerberProtelExtensions"])
    setExcludeEdgeLayer(popt, settings["ExcludeEdgeLayer"])
    popt.SetScale(1)
    popt.SetUseAuxOrigin(settings["UseAuxOrigin"])
    popt.SetUseGerberX2format(False)

    # This by gerbers only
    popt.SetSubtractMaskFromSilk(False)
    popt.SetDrillMarksType(pcbnew.DRILL_MARKS_NO_DRILL_SHAPE)
    popt.SetSkipPlotNPTH_Pads(False)

    # prepare the gerber job file
    jobfile_writer = GERBER_JOBFILE_WRITER(board)

    for name, id, comment in plot_plan:
        if id <= B_Cu:
            popt.SetSkipPlotNPTH_Pads(True)
        else:
            popt.SetSkipPlotNPTH_Pads(False)

        pctl.SetLayer(id)
        suffix = "" if settings["NoSuffix"] else name
        pctl.OpenPlotfile(suffix, PLOT_FORMAT_GERBER, comment)
        jobfile_writer.AddGbrFile(id, os.path.basename(pctl.GetPlotFileName()))
        if pctl.PlotLayer() == False:
            raise RuntimeError("KiCAD plot error")

    if hasCopper(plot_plan):
        #generate internal copper layers, if any
        for i, layer in enumerate(Layer.innerCu(board.GetCopperLayerCount())):
            popt.SetSkipPlotNPTH_Pads(True)
            pctl.SetLayer(layer)
            layerName = "" if settings["NoSuffix"] else f"inner{i + 1}"
            pctl.OpenPlotfile(layerName, PLOT_FORMAT_GERBER, "inner")
            jobfile_writer.AddGbrFile(layer, os.path.basename(pctl.GetPlotFileName()))
            if pctl.PlotLayer() == False:
                raise RuntimeError("KiCAD plot error")

    # At the end you have to close the last plot, otherwise you don't know when
    # the object will be recycled!
    pctl.ClosePlot()

    if drilling:
        # Fabricators need drill files.
        # sometimes a drill map file is asked (for verification purpose)
        drlwriter = EXCELLON_WRITER(board)
        mapFmt = settings['MapFileFormat']
        drlwriter.SetMapFileFormat(mapFmt)

        mirror = False
        minimalHeader = settings["MinimalHeader"]
        if settings["UseAuxOrigin"]:
            offset = board.GetDesignSettings().GetAuxOrigin()
        else:
            offset = VECTOR2I(0, 0)

        # False to generate 2 separate drill files (one for plated holes, one for non plated holes)
        # True to generate only one drill file
        mergeNPTH = settings["MergeNPTH"]
        drlwriter.SetOptions(mirror, minimalHeader, offset, mergeNPTH)
        drlwriter.SetRouteModeForOvalHoles(False)

        metricFmt = True
        zerosFmt = settings["ZerosFormat"]
        drlwriter.SetFormat(metricFmt, zerosFmt)
        genDrl = True
        genMap = True
        drlwriter.CreateDrillandMapFilesSet(pctl.GetPlotDirName(), genDrl, genMap)

        # One can create a text file to report drill statistics
        rptfn = pctl.GetPlotDirName() + 'drill_report.rpt'
        drlwriter.GenDrillReportFile(rptfn)

    job_fn=os.path.dirname(pctl.GetPlotFileName()) + '/' + os.path.basename(boardfile)
    job_fn=os.path.splitext(job_fn)[0] + '.gbrjob'
    jobfile_writer.CreateJobFile(job_fn)

def pasteDxfExport(board, plotDir):
    pctl = PLOT_CONTROLLER(board)
    popt = pctl.GetPlotOptions()

    popt.SetOutputDirectory(os.path.abspath(plotDir))
    popt.SetAutoScale(False)
    popt.SetScale(1)
    popt.SetMirror(False)
    setExcludeEdgeLayer(popt, True)
    popt.SetScale(1)
    popt.SetDXFPlotUnits(DXF_UNITS_MM)
    popt.SetDXFPlotPolygonMode(False)
    popt.SetDrillMarksType(DRILL_MARKS_NO_DRILL_SHAPE)

    plot_plan = [
        # name, id, comment
        ("PasteBottom", B_Paste, "Paste Bottom"),
        ("PasteTop", F_Paste, "Paste top"),
        ("EdgeCuts", Edge_Cuts, "Edges"),
    ]

    output = []
    for name, id, comment in plot_plan:
        pctl.SetLayer(id)
        pctl.OpenPlotfile(name, PLOT_FORMAT_DXF, comment)
        output.append(pctl.GetPlotFileName())
        if pctl.PlotLayer() == False:
            print("plot error")
    pctl.ClosePlot()
    return tuple(output)

def dxfImpl(boardfile, outputdir):
    basename = os.path.dirname(boardfile)
    if outputdir:
        plotDir = outputdir
    else:
        plotDir = basename
    plotDir = os.path.abspath(plotDir)

    board = LoadBoard(boardfile)

    pasteDxfExport(board, plotDir)
